# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: row_handler.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='row_handler.proto',
  package='bytegain.custom.model_data.row_handler_proto',
  syntax='proto2',
  serialized_pb=_b('\n\x11row_handler.proto\x12,bytegain.custom.model_data.row_handler_proto\"R\n\nRowHandler\x12\x44\n\x06\x62ucket\x18\x01 \x03(\x0b\x32\x34.bytegain.custom.model_data.row_handler_proto.Bucket\"\x9e\x03\n\x06\x42ucket\x12\x0c\n\x04name\x18\x01 \x01(\t\x12T\n\x0b\x62ucket_type\x18\x02 \x01(\x0e\x32?.bytegain.custom.model_data.row_handler_proto.Bucket.BucketType\x12J\n\textractor\x18\x03 \x01(\x0b\x32\x37.bytegain.custom.model_data.row_handler_proto.Extractor\x12\x10\n\x08has_null\x18\x04 \x01(\x08\x12\x14\n\x0c\x62ucket_count\x18\x05 \x01(\x05\x12\x15\n\rfloat_buckets\x18\x06 \x03(\x02\x12\x16\n\x0estring_buckets\x18\x07 \x03(\t\x12\x13\n\x0bint_buckets\x18\x08 \x03(\x05\x12\x17\n\x0f\x62ucket_fraction\x18\t \x03(\x02\x12\x12\n\nnone_value\x18\n \x01(\x02\"K\n\nBucketType\x12\n\n\x06\x44IRECT\x10\x00\x12\n\n\x06STRING\x10\x01\x12\n\n\x06\x42UCKET\x10\x02\x12\x0b\n\x07\x42OOLEAN\x10\x03\x12\x0c\n\x08\x43\x41TEGORY\x10\x04\"\x86\x04\n\tExtractor\x12]\n\x0e\x65xtractor_type\x18\x01 \x01(\x0e\x32\x45.bytegain.custom.model_data.row_handler_proto.Extractor.ExtractorType\x12\x12\n\nfield_name\x18\x02 \x01(\t\x12\n\n\x02mu\x18\x03 \x01(\x02\x12\r\n\x05sigma\x18\x04 \x01(\x02\x12\x14\n\x0c\x66ield_name_2\x18\x05 \x01(\t\x12\x0f\n\x07min_val\x18\x06 \x01(\x02\x12\x0f\n\x07max_val\x18\x07 \x01(\x02\x12\x13\n\x0bpercentiles\x18\x08 \x03(\x02\x12S\n\tdata_type\x18\t \x01(\x0e\x32@.bytegain.custom.model_data.row_handler_proto.Extractor.DataType\x12\x15\n\rnull_fraction\x18\n \x01(\x02\"|\n\rExtractorType\x12\n\n\x06\x44IRECT\x10\x00\x12\r\n\tAUTO_NORM\x10\x01\x12\x0c\n\x08NON_NULL\x10\x02\x12\x0c\n\x08LOG_NORM\x10\x03\x12\x13\n\x0fPERCENTILE_NORM\x10\x04\x12\t\n\x05RATIO\x10\x05\x12\x14\n\x10\x44\x45\x46\x41ULTED_DIRECT\x10\x06\"4\n\x08\x44\x61taType\x12\x07\n\x03INT\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\n\n\x06STRING\x10\x02\x12\x08\n\x04\x42OOL\x10\x03')
)



_BUCKET_BUCKETTYPE = _descriptor.EnumDescriptor(
  name='BucketType',
  full_name='bytegain.custom.model_data.row_handler_proto.Bucket.BucketType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DIRECT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BUCKET', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CATEGORY', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=491,
  serialized_end=566,
)
_sym_db.RegisterEnumDescriptor(_BUCKET_BUCKETTYPE)

_EXTRACTOR_EXTRACTORTYPE = _descriptor.EnumDescriptor(
  name='ExtractorType',
  full_name='bytegain.custom.model_data.row_handler_proto.Extractor.ExtractorType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DIRECT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AUTO_NORM', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NON_NULL', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG_NORM', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PERCENTILE_NORM', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RATIO', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEFAULTED_DIRECT', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=909,
  serialized_end=1033,
)
_sym_db.RegisterEnumDescriptor(_EXTRACTOR_EXTRACTORTYPE)

_EXTRACTOR_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='bytegain.custom.model_data.row_handler_proto.Extractor.DataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOL', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1035,
  serialized_end=1087,
)
_sym_db.RegisterEnumDescriptor(_EXTRACTOR_DATATYPE)


_ROWHANDLER = _descriptor.Descriptor(
  name='RowHandler',
  full_name='bytegain.custom.model_data.row_handler_proto.RowHandler',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bucket', full_name='bytegain.custom.model_data.row_handler_proto.RowHandler.bucket', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=67,
  serialized_end=149,
)


_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='bytegain.custom.model_data.row_handler_proto.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_type', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.bucket_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extractor', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.extractor', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_null', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.has_null', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_count', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.bucket_count', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='float_buckets', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.float_buckets', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_buckets', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.string_buckets', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int_buckets', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.int_buckets', index=7,
      number=8, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_fraction', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.bucket_fraction', index=8,
      number=9, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='none_value', full_name='bytegain.custom.model_data.row_handler_proto.Bucket.none_value', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BUCKET_BUCKETTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=566,
)


_EXTRACTOR = _descriptor.Descriptor(
  name='Extractor',
  full_name='bytegain.custom.model_data.row_handler_proto.Extractor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='extractor_type', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.extractor_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='field_name', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.field_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mu', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.mu', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sigma', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.sigma', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='field_name_2', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.field_name_2', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_val', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.min_val', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_val', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.max_val', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='percentiles', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.percentiles', index=7,
      number=8, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.data_type', index=8,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='null_fraction', full_name='bytegain.custom.model_data.row_handler_proto.Extractor.null_fraction', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _EXTRACTOR_EXTRACTORTYPE,
    _EXTRACTOR_DATATYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=569,
  serialized_end=1087,
)

_ROWHANDLER.fields_by_name['bucket'].message_type = _BUCKET
_BUCKET.fields_by_name['bucket_type'].enum_type = _BUCKET_BUCKETTYPE
_BUCKET.fields_by_name['extractor'].message_type = _EXTRACTOR
_BUCKET_BUCKETTYPE.containing_type = _BUCKET
_EXTRACTOR.fields_by_name['extractor_type'].enum_type = _EXTRACTOR_EXTRACTORTYPE
_EXTRACTOR.fields_by_name['data_type'].enum_type = _EXTRACTOR_DATATYPE
_EXTRACTOR_EXTRACTORTYPE.containing_type = _EXTRACTOR
_EXTRACTOR_DATATYPE.containing_type = _EXTRACTOR
DESCRIPTOR.message_types_by_name['RowHandler'] = _ROWHANDLER
DESCRIPTOR.message_types_by_name['Bucket'] = _BUCKET
DESCRIPTOR.message_types_by_name['Extractor'] = _EXTRACTOR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RowHandler = _reflection.GeneratedProtocolMessageType('RowHandler', (_message.Message,), dict(
  DESCRIPTOR = _ROWHANDLER,
  __module__ = 'row_handler_pb2'
  # @@protoc_insertion_point(class_scope:bytegain.custom.model_data.row_handler_proto.RowHandler)
  ))
_sym_db.RegisterMessage(RowHandler)

Bucket = _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), dict(
  DESCRIPTOR = _BUCKET,
  __module__ = 'row_handler_pb2'
  # @@protoc_insertion_point(class_scope:bytegain.custom.model_data.row_handler_proto.Bucket)
  ))
_sym_db.RegisterMessage(Bucket)

Extractor = _reflection.GeneratedProtocolMessageType('Extractor', (_message.Message,), dict(
  DESCRIPTOR = _EXTRACTOR,
  __module__ = 'row_handler_pb2'
  # @@protoc_insertion_point(class_scope:bytegain.custom.model_data.row_handler_proto.Extractor)
  ))
_sym_db.RegisterMessage(Extractor)


# @@protoc_insertion_point(module_scope)



texts = {'Accept and Close', '', 'Next', 'Skip this question', 'See my matches', 'See More Matches'}
skip_text = {'', 'Next'}


def rows_filter(rows):
    """
    Keep pages and some clicks. Remove clicks made during quiz (they have either empty text or "Next").
    :param rows: user session
    :return: filtered user session
    """
    rows = [row for row in rows
            if (
                row.get('type', '') == 'page'
                or (row.get('event', '') in ('BG Click', 'BG Click2')
                    and row.get('bytegain', {}).get('clickText', '').strip() not in skip_text)
            )
            ]
    return rows


import time
import sys
import ujson
import gzip
import boto3
import os
from botocore.exceptions import ClientError
from google.cloud import storage
from apiclient import discovery, http
from googleapiclient.errors import HttpError
import argparse
import tempfile
import bytegain.data_import.segment_utils as segment_utils
from bytegain.custom.data_import.schema_info import TableInfo, ColumnInfo, SchemaInfo, save_schema

from bytegain.data_import.segment_schema_filter import SchemaFilter
from random import Random
from collections import defaultdict
import queue
import threading
from numbers import Number

class ColumnTracker(object):
    """A ColumnTracker tracks information about a column and the values it contains.
    The number of occurrences of the first _max_items distinct values are counted.
    """
    def __init__(self, name, max_items):
        self._name = name
        self._max_items = max_items
        self._counts = defaultdict(int)
        self._total_non_null_count = 0
        self._total_count = 0
        self._datatype = None

    def add_entry(self, value):
        try:
            if self._datatype is None and value is not None:
                py_type = type(value)
                if py_type == str or py_type == str:
                    self._datatype = "varchar"
                elif py_type == bool or py_type == int:
                    self._datatype = "int"
                elif isinstance(value, Number):
                    self._datatype = "float"
                else:
                    self._datatype = str(py_type)

            if value in self._counts or len(self._counts) < self._max_items:
                self._counts[value] += 1

            self._total_count += 1
            if value is not None:
                self._total_non_null_count += 1
        except Exception as e:
            print("Could not add %s to %s: %s" % (str(value), self._name, str(e)))
            pass

    def get_top_values(self, min_percent):
        values = []
        min_count = self._total_count * min_percent
        for key, value in sorted(iter(list(self._counts.items())), key=lambda k_v: (k_v[1],k_v[0]), reverse = True):
            if value >= min_count:
                values.append((key, value * 100.0 / self._total_count))

        return values

class TableTracker(object):
    """A TableTracker is a collection of ColumnTrackers for each column in a table.
    """
    def __init__(self, name, max_column_items, is_tracks = False):
        self._name = name
        self._max_column_items = max_column_items
        self._columns = {}
        self._total_count = 0
        self._no_userid  = 0
        self._is_tracks = is_tracks

    def _add_column(self, column_name, value):
        if column_name not in self._columns:
            self._columns[column_name] = ColumnTracker(column_name, self._max_column_items)
        self._columns[column_name].add_entry(str(value, "UTF-8") if type(value) == str else value)

    def _add_columns(self, iteritems, prefix = None):
        for column_name, value in iteritems:
            full_name = ("%s.%s" % (prefix, column_name)) if prefix else column_name
            if type(value) == dict:
                self._add_columns(iter(list(value.items())), prefix = full_name)
            elif type(value) in [list, tuple]:
                for entry in value:
                  self._add_columns([(column_name, entry)], prefix = prefix)                    
            else:
                self._add_column(full_name, value)

    def add_entry(self, values, has_userid):
        self._total_count += 1
        if not has_userid:
            self._no_userid += 1

        if self._is_tracks:
            self._add_columns([("event", values["event"])])
            self._add_columns(iter(list(values["context"].items())))
        else:
            self._add_columns(iter(list(values.items())))

    def print_pretty(self, min_percent):
        print("%s: %d [user_id: %3f%%]" % (self._name, self._total_count, 100 - (self._no_userid * 100.0 / self._total_count)))
        for key, column in sorted(iter(list(self._columns.items())), key=lambda k_v1: (k_v1[1]._total_count, k_v1[0]), reverse = True):
            if column._datatype is not None:
                if column._total_count > self._total_count * 0.02:
                    top_values = column.get_top_values(min_percent)
                    top_values = ["%s: %d%%" % (x[0], x[1]) for x in top_values]
                    print("    %14s: %s" % (key, str(top_values)))

class WorkerThread(threading.Thread):
    def __init__(self, file_queue, ignored_events):
        super(WorkerThread, self).__init__()
        self._file_queue = file_queue
        self._ignored_events = ignored_events
        self._event_types = defaultdict(int)
        self._page_table = TableTracker("page", 1000)
        self._tracks_table = TableTracker("tracks", 1000, is_tracks=True)
        self._other_count = defaultdict(int)
        self._event_tables = {}
        self._track_count = 0
        self._total_time = 0
        self._files = 0
        self._ignored_event_count = 0
        self._start_time = time.mktime(args.start_date.utctimetuple()) if args.start_date else 0
        self._end_time = time.mktime(args.end_date.utctimetuple()) if args.end_date else sys.maxsize

        print("Ignored events: %s" % str(self._ignored_events))

    def total_count(self):
        return self._tracks_table._total_count

    def run(self):
        ended = False
        while not ended:
            item = self._file_queue.get()
            if not item:
                ended = True
            else:
                self.process_file(item)
            self._file_queue.task_done()

    def process_file(self, file_name):
        start_time = time.time()
        input_file = gzip.open(file_name, "rb")
        for i, line in enumerate(input_file):
            content = ujson.loads(line)
            if "receivedAt" in content:
                # Remove trailing Z
                date = content["receivedAt"][0:-1]
                epoch = segment_utils.timestamp_to_epoch(date)
                if epoch < self._start_time or epoch > self._end_time:
                    break
            has_userid = "userId" in content
            if args.require_user_id and not has_userid:
                continue
            content_type = content["type"]
            self._event_types[content_type] += 1
            if content_type == "track":
                event_type = content["event"]
                if event_type in self._ignored_events:
                    self._ignored_event_count +=1
                    continue

                self._tracks_table.add_entry(content, has_userid)

                self._track_count += 1
                if event_type not in self._event_tables:
                    self._event_tables[event_type] = TableTracker(event_type, 1000)

                event_table = self._event_tables[event_type]
                if "properties" in content:
                    properties = content["properties"]
                    event_table.add_entry(properties, has_userid)
            elif content_type == "page":
                self._page_table.add_entry(content["properties"], has_userid)
            else:
                self._other_count[content_type] += 1

        self._total_time += (time.time() - start_time)
        self._files += 1
        os.remove(file_name)

def create_table_info(schema, table_tracker):
    table_info = TableInfo(schema, table_tracker._name)
    table_info.set_rows(table_tracker._total_count)
    table_info.add_column(ColumnInfo("userId", "varchar"))
    table_info.add_column(ColumnInfo("receivedAt", "timestamp"))

    for column_tracker in list(table_tracker._columns.values()):
        if column_tracker._datatype is None:
            continue

        name = column_tracker._name
        null_fraction = 1.0 - float(column_tracker._total_non_null_count) / table_tracker._total_count
        column = ColumnInfo(name, column_tracker._datatype)
        column.null_fraction = null_fraction

        top_values = column_tracker.get_top_values(0.00001)
        if len(top_values) > 0 and top_values[0] > 100:
            for value in top_values:
                column.add_group(value[0], value[1] / 100.0)

        table_info.add_column(column)
    return table_info

class S3Iterator(object):
    def __init__(self, s3_url):
        session = boto3.Session(profile_name=args.s3_profile)
        self._s3 = session.client('s3')
        self._bucket, self._key, _ = segment_utils.parse_s3_url(s3_url)
        self._file_iterator = None
        self._current_key = self._key

    def get_file(self, key):
        save_file = tempfile.NamedTemporaryFile(delete = False)
        try:
            self._s3.download_file(self._bucket, key, save_file.name)
            return save_file.name            
        except ClientError as e:
            print("Could not access: %s" % key)
            return None

    def __iter__(self):
        return self

    def _refresh_iterator(self):
        response = s3.list_objects_v2(Bucket=self._bucket, Prefix=self._key, StartAfter = self._current_key)
        # print "response: %s" % str(response)
        if "Contents" in response:
            contents = response["Contents"]
            print("Contents len: %d" % len(contents))
            self._file_iterator = contents.__iter__()

    def __next__(self):
        if self._file_iterator == None: 
            self._refresh_iterator()

        if self._file_iterator == None:
            raise StopIteration()

        try:           
            self._current_key = self._file_iterator.next()["Key"]
            return self._current_key
        except StopIteration:
            self._file_iterator = None
            return next(self)

class GSIterator(object):
    def __init__(self, gs_url):        
        gs_url = gs_url.replace('*', '')
        self._service = discovery.build('storage', 'v1')
        self._bucket_name, self._key, _ = segment_utils.parse_s3_url(gs_url)
        self._client = storage.Client()
        self._bucket = self._client.get_bucket(self._bucket_name)
        self._file_iterator = None
        self._page_token = None
        self._is_done = False

    def get_file(self, key, retry = True):
        with tempfile.NamedTemporaryFile(delete = False) as gz_file: 
            print("loading: %s" % key)
            try:
                req = self._service.objects().get_media(bucket=self._bucket_name, object=key)   
                downloader = http.MediaIoBaseDownload(gz_file, req)
                done = False
                while done is False:
                    status, done = downloader.next_chunk()
                return gz_file.name
            except HttpError as e:
                if retry:
                    print("Error reading: %s.  Retrying" % key)
                    return get_file(key, False)
                else:
                    raise e

    def __iter__(self):
        return self

    def _refresh_iterator(self):
        print("Listing: %s" % self._key)
        if self._is_done:
            raise StopIteration
        blobs = self._bucket.list_blobs(prefix=self._key, page_token = self._page_token)
        self._page_token = blobs.next_page_token
        if self._page_token == None:
            self._is_done = True
        self._file_iterator = blobs.__iter__()

    def __next__(self):
        if self._file_iterator == None: 
            self._refresh_iterator()

        try:           
            blob = next(self._file_iterator)
            self._current_key = blob.name
            return self._current_key
        except StopIteration:
            self._refresh_iterator()            
            return next(self)

parser = argparse.ArgumentParser(description='Analyzes Segment JSON data in S3')
parser.add_argument('--s3_src', type=str, help='s3 path of data ', nargs = '+', required=True)
parser.add_argument('--s3_profile', type=str, help='s3 profile to use for data access', default = "s3_data_user")
parser.add_argument('--sample_ratio', type=float, help='fraction of data to read', default = 0.01)
parser.add_argument('--start_date', type = segment_utils.as_date, help = 'Start date of data to analyze', default = None)
parser.add_argument('--end_date', type = segment_utils.as_date, help = 'End date of data to analyze', default = None)
parser.add_argument('--sampling_seed', type = int, help = 'Random seed for sampling', default = 91876)
parser.add_argument('--max_files', type = int, help = 'Maximum files to sample', default = None)
parser.add_argument('--schema_info', type = str, help = 'SchemaInfo file to create', required = True)
parser.add_argument('--filter_config', type=str, help='JSON filter config file', required=True)
parser.add_argument('--client', type = str, help = 'Client name', required = True)
parser.add_argument('--require_user_id', type = bool, help = "If true, then only count rows with userId set", default = False)

args = parser.parse_args()
rng = Random()
rng.seed(args.sampling_seed)

filter_config = SchemaFilter.load_from_json(args.filter_config)

file_queue = queue.Queue(20)

thread = WorkerThread(file_queue, filter_config.ignored_events)
thread.daemon = True
thread.start()
file_count = 0
start_time = time.time()
       
for s3_path in args.s3_src:
    print("Processing: '%s'" % s3_path)
    files = S3Iterator(s3_path) if "s3://" in s3_path else GSIterator(s3_path)
    for start_key in files:
        if args.max_files != None and file_count >= args.max_files:
            break

        if rng.random() < args.sample_ratio:
            file_name = files.get_file(start_key)
            if file_name is not None:
                file_queue.put(file_name)
                file_count += 1
                if file_count % 10 == 0:
                    total_time = time.time() - start_time
                    sys.stdout.write("\r%d files in %ds (%5.3f/s)         " % (file_count, total_time, file_count / total_time))
                    sys.stdout.flush()

file_queue.put(False)
file_queue.join()
print("Other types: %s" % str(thread._other_count))

print(thread._page_table.print_pretty(0.02))
print(thread._tracks_table.print_pretty(0.02))

# for table in sorted(thread._event_tables.values(), key = lambda (t): t._total_count, reverse = True):
#       if table._total_count > thread.total_count() * args.min_event_fraction:
#               print table.print_pretty(0.02)
#       else:
#               print "Skipping %s.  Too small %5.4f%%" % (table._name, table._total_count * 100.0 / thread.total_count())

track_info = create_table_info(args.client, thread._tracks_table)
page_info = create_table_info(args.client, thread._page_table)

# First pass, take filter min and use threshold 20x lower
min_event_fraction = filter_config.min_join_table_row_fraction / 20
join_tables = {}
for table_tracker in list(thread._event_tables.values()):
    if table_tracker._total_count > thread.total_count() * min_event_fraction:
        join_tables[table_tracker._name] = create_table_info(args.client, table_tracker)

schema_info = SchemaInfo(args.client, track_info, page_info, join_tables)

save_schema(args.schema_info, schema_info)



import pyximport
pyximport.install()
import os
import sys
import numpy
import importlib
import bytegain.custom.nn_tools.xgb_classifier as xgb
import bytegain.custom.nn_tools.model_utils as model_utils
import bytegain.custom.model_data.feature_analysis as feature_analysis
import bytegain.nn_tools.auc_plot as auc_plot
from bytegain.custom.nn_tools.prod_classifier import ProdXgbClassifier
from bytegain.custom.nn_tools.xgb_classifier import XgbClassifier
import bytegain.custom.model_data.read_from_db as read_from_db
from bytegain.custom.model_data.row_handler import RowHandler
from bytegain.custom.model_data.bucket import ValueBucket
from bytegain.custom.model_data.extractor import RowExtractor
from bytegain.custom.cust.apartmentlist.feature_filter import FeatureFilter

"""
import bytegain.cust.apartmentlist.al_xgb as al_xgb
import bytegain.nn_tools.model_utils as model_utils
import bytegain.nn_tools.xgb_classifier as xgb
rh, datasets = al_xgb.create_rh_and_dataset('al_handler_test.json')
results = xgb.cv(datasets, xgb.XgbClassifier, 401, parallel=True)

import bytegain.cust.apartmentlist.al_xgb as al_xgb
import bytegain.nn_tools.xgb_classifier as xgb
rh, datasets = al_xgb.create_rh_and_dataset('al_features_20161214.json')
model = xgb.XgbClassifier()
model.train(datasets)


import bytegain.nn_tools.knock_out as knock_out
import bytegain.cust.apartmentlist.leads_from_redshift as tfl
import bytegain.nn_tools.xgb_classifier as xgb

features = al_handler_from_features.read_al_features('al_handler1.json')
rh = al_handler_from_features.create_row_handler(features)
datasets = tfl.create_datasets_from_row_handler(rh, test_fraction = 0.2)
results = xgb_classifier.cv(datasets, xgb_classifier.XgbClassifier, 601)


"""

def output_feature_scores(model, rh, output):
    features = model.feature_scores(rh)

    for feature in features:
        output.write("%25s: %4.1f\n" % (feature[0], feature[1]));


def train_from_dataset(config, rh, dataset):
    model = xgb.XgbClassifier()
    model.train(dataset, min_child_weight=config['min_child_weight'], max_rounds=config['max_rounds'],
                max_depth=config['max_depth'], l2_norm = config['l2_norm'], train_step_size = config['train_step_size'])
    control_results = model.get_sorted_results(dataset.get_test_control())
    print("Control AUC: %5.3f" % auc_plot.compute_auc(control_results))

    output_feature_scores(model, rh, sys.stdout)
    return model

def create_rh_and_dataset(config):
    rh = create_row_handler(config)
    datasets = create_datasets_from_row_handler(rh, config, test_fraction = 0.2)
    return rh, datasets

def train_and_push(config):
    model_name = config['model_name']
    if not os.path.isdir(model_name):
        print("Creating %s" % model_name)
        os.mkdir(model_name)

    rh, datasets = create_rh_and_dataset(config)
    classifier = train_from_dataset(config, rh, datasets)
    results = classifier.get_sorted_results(datasets.test)

    if not os.path.isdir(model_name):
        os.mkdir(model_name)
    os.chdir(model_name)
    model_name = "%s_%s" % (config['model_name'], config['model_version'])
    try:
        auc_plot.plot_predicted_vs_actual2('%s_p_vs_a.png' % model_name, results, points=20)
        auc_plot.plot_auc('%s_roc.png' % model_name, results)

        spec = model_utils.XgbModelSpec(model_name, ".")
        spec.save(classifier, datasets._row_handler, results)
        model_utils.save_results_as_csv(model_name + '.csv', 'id', 'pred', results)
        spec.upload()
    finally:
        os.chdir('..')


def get_selector(config):
    if config['downsample']:
        where = '(%s OR CORE.public.f_bucket_number(%s, 1000) < 50)' % (config['outcome_field'], config['id_field'])
    else:
        where = None

    return read_from_db.Selector(config['table'], where,
                                 config['outcome_field'],
                                 config['id_field'],
                                 None,
                                 config.get('positive_outcome_rate_field'))

def create_datasets_from_row_handler(row_handler, config, test_fraction = 0.2, rows = None):
    selector = get_selector(config)
    numpy.random.seed(config['random_seed'])
    read_from_db.generate_row_handler_stats(selector, row_handler)
    return read_from_db.create_datasets(selector, row_handler, test_fraction, rows)

def get_feature_json_file(config):
    model_name = config['model_name']
    if not os.path.isdir(model_name):
        os.mkdir(model_name)
    return "%s/%s" % (model_name, config['feature_json'])

def run_feature_analysis(config):
    data_info = feature_analysis.DataInfo(get_selector(config))
    features = feature_analysis.process_features(data_info)
    feature_analysis.save_features(config['feature_json'], features)
    return features


def create_row_handler(config, load_from = None):
    if isinstance(config['filter'], FeatureFilter):
        filter = config['filter']
    else:
        # For backwards compatibility let the filter be in another file loaded by name.
        module = importlib.import_module('bytegain.custom.cust.apartmentlist.%s' % config['filter'])
        filter = getattr(module, 'filter')

    features = feature_analysis.read_features(config['feature_json'])

    # Set this to experiment with weighting.
    if config['control_field']:
        print("Using '%s' as control.  Weight: %d" % (config['control_field'], config['control_weight']))
        weight_extractor = ControlWeightExtractor(config['control_field'], config['control_weight'])
        control_extractor = RowExtractor(config['control_field'])
    else:
        weight_extractor = None
        control_extractor = None

    row_handler = RowHandler(label_bucket = ValueBucket("leased", RowExtractor(config['outcome_field']), has_null = False),
                             id_extractor = RowExtractor(config['id_field']),
                             weight_extractor = weight_extractor,
                             control_extractor = control_extractor,
                             load_from = load_from,
                             extra_fields = [])
    if load_from is None:
        # print "Excluded features: %s" % str(EXCLUDED_FEATURES)
        buckets = filter.create_buckets(features)
        for bucket in buckets:
            row_handler.add_bucket(bucket)

    return row_handler


class ControlWeightExtractor(object):
    def __init__(self, control_field, control_weight = 1.0):
        self._control_weight = control_weight
        self._control_field = control_field
        self._field = None

    def get_fields(self):
        return [self._control_field]

    def update_fields_by_id(self, cursor_header):
        self._field = cursor_header.index(self._control_field)

    def extract(self, row):
        if self._field is not None:
            return self._control_weight if row[self._field] else 1.0
        else:
            return self._control_weight if row[self._control_field] else 1.0

    def description(self):
        return "Control weighting: %f" % self._control_weight



"""
Typical usage:

python clickstream_to_proto.py --config model_configs/al_1.0 --output /data3/al_data/clickstream-al_1.0

Testing:

python clickstream_to_proto.py --config model_configs/al_1.0  --output protos --shards 1 --workers 1 --row_limit 100

"""

import argparse
import os

from bytegain.custom.model_data.read_from_db import Selector, write_protos
from bytegain.custom.model_data.model_config import ModelConfig, ModelConfigRowHandler

parser = argparse.ArgumentParser(description='Write clickstream protos.')
parser.add_argument('--config', type=str, required=True, help='Path to model config')
parser.add_argument('--output', type=str, required=True, help='Output filebase')
parser.add_argument('--shards', type=int, default=1024, help='Output shard count')
parser.add_argument('--workers', type=int, default=8, help='Worker count')
parser.add_argument('--row_limit', type=int, default=None, help='Maximum number of rows to process per worker')
args = parser.parse_args()

config = ModelConfig.from_json_file(args.config)
row_handler = ModelConfigRowHandler(config)

for worker in range(args.workers):
    if args.workers == 1 or os.fork() == 0:
        # Apportion the shards sequentially among the workers.
        # This is going to be most efficient when args.workers divides evenly into shard_count.
        stride = args.shards / args.workers
        shard_range = [worker * stride, (worker + 1) * stride]

        where = "(%s) and strtol(right(md5(%s), 15), 16) %% %d >= %d and strtol(right(md5(%s), 15), 16) %% %d < %d" % (
                config.where, config.id_field, args.shards, shard_range[0], config.id_field, args.shards, shard_range[1])
        # TODO(chris): Upgrade to fully-qualified table name
        # TODO(chris): Stop peeking into row_handler
        selector = Selector(config.table, where, row_handler._label_extractor._field_name, config.id_field)
        write_protos(selector, row_handler, args.output, shard_range, args.shards, row_limit=args.row_limit)
        exit()

for worker in range(args.workers):
    os.wait()



import argparse
from datetime import datetime, timedelta

from google.cloud import storage

import bytegain.custom.model_data.feature_analysis as model_fa
from . import feature_analysis
from .compare_features import compare_features


def features_filename(features_type, date_str):
    return '%s/%s-features-%s.json' % (date_str, features_type, date_str.replace('/', '-'))


def date_str(d):
    return d.strftime("%Y/%m/%d")


parser = argparse.ArgumentParser(description='Analyze features from AL request logs and optionally notify.')
parser.add_argument('--type', help='Type of features: lease, search_rank',
                    choices=[feature_analysis.TYPE_LEASE, feature_analysis.TYPE_SEARCH_RANK], required=True)
args = parser.parse_args()

gcs_client = storage.Client(project='bg-rest-api')
bucket = gcs_client.lookup_bucket('al-log-analysis')

today = date_str(datetime.now())

print('Running feature analysis for requests from %s (UTC)' % today)
where = "_PARTITIONTIME = TIMESTAMP('%s')" % today.replace('/', '-')
if today == date_str(datetime.utcnow()):
    where = '(' + where + ' OR _PARTITIONTIME IS NULL)'
stats_runner = feature_analysis.StatsRunner(feature_analysis.rows_from_bq_query(args.type, where))
current_features = stats_runner.run()
current_features_json = model_fa.features_to_json(current_features)

current_blob = bucket.blob(features_filename(args.type, today))
print('Uploading current features to: gs://%s/%s' % (current_blob.bucket.name, current_blob.name))
current_blob.upload_from_string(current_features_json, content_type='application/json')

previous_feature_analysis_json = None
previous_blob = None
previous_date = None
# Go back 10 days and try and find something if yesterday's run failed to produce a file.
for i in range(10):
    previous_date = date_str(datetime.now() - timedelta(days=i + 1))
    previous_blob = bucket.get_blob(features_filename(args.type, previous_date))
    if previous_blob:
        previous_feature_analysis_json = previous_blob.download_as_string()
        break

if not previous_feature_analysis_json:
    raise Exception('Cannot find previous feature analysis file in GCS')

print('Previous features: gs://%s/%s' % (previous_blob.bucket.name, previous_blob.name))

previous_features = model_fa.features_from_json(previous_feature_analysis_json)

if compare_features(current_features, today, previous_features, previous_date):
    exit(1)



import argparse
import os
import subprocess
import time
import uuid
from sys import stdout

from google.cloud import bigquery

client = bigquery.Client()


def async_query(query):
    print('\n>>> Running async query ...')

    job = client.run_async_query(uuid.uuid4().hex, query)
    job.use_legacy_sql = False
    job.use_query_cache = True

    job.result()  # Wait for job to complete.

    if job.error_result or job.errors:
        print('\n!!! Query job finished with errors:\n', job.error_result, '\n', job.errors)
        exit(1)

    print('Query returned %d rows.' % job.query_results().total_rows)

    job.destination.reload()

    return job.destination


def export_table(table, gcs_pattern):
    print('\n>>> Extracting table to GCS ...')
    job = client.extract_table_to_storage(uuid.uuid4().hex, table, gcs_pattern)
    job.destination_format = 'NEWLINE_DELIMITED_JSON'
    job.compression = 'GZIP'

    job.begin()

    # Wait for job to complete
    retry_count = 100
    interval = 3  # seconds
    stdout.write('Waiting for job to complete ')
    while retry_count > 0 and job.state != 'DONE':
        retry_count -= 1
        stdout.write('.')
        stdout.flush()
        time.sleep(interval)
        job.reload()

        if job.error_result or job.errors:
            print('\n\n!!! Extract job finished with errors:\n', job.error_result, '\n', job.errors)
            exit(1)
    if retry_count == 0:
        print('\n\n!!! Extract job taking too long. Bailing. The job may eventually complete, you\'re on your own!\n')
        exit(1)
    print('.. DONE')


def gcs_cleanup(gcs_pattern):
    print('\n>>> Cleaning up GCS %s ...' % gcs_pattern)
    try:
        # quick and dirty gsutil call so we don't have to implement it all using the low level storage lib
        print(subprocess.check_output(['gsutil', '-m', 'rm', gcs_pattern], stderr=subprocess.STDOUT))
    except subprocess.CalledProcessError as cpe:
        if '1 files/objects could not be removed' in cpe.output:
            # try again without the -m flag due to some weirdness in gsutil
            try:
                print(subprocess.check_output(['gsutil', 'rm', gcs_pattern], stderr=subprocess.STDOUT))
            except subprocess.CalledProcessError as cpe2:
                if 'No URLs matched' not in cpe2.output:
                    print(cpe2.output)
                    raise cpe2
        else:
            print(cpe.output)
            raise cpe


def local_copy(gcs_pattern, local_dir):
    print('\n>>> Copying from GCS to local ...')
    # quick and dirty gsutil call so we don't have to implement it all using the low level storage lib
    print(subprocess.check_output(['gsutil', '-m', 'cp', gcs_pattern, local_dir], stderr=subprocess.STDOUT))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Export the results of a BigQuery query to GCS and optionally copy the files locally.')
    parser.add_argument('query',
                        help='the query to run. For complex queries use a file as input and pass "$(<query_file)", including the quotes.')
    parser.add_argument('gcs_prefix', help='the GCS URL prefix, e.g. "gs://bucket/dir/prefix"')
    parser.add_argument('--local-copy', dest='dir',
                        help='make a local copy in the target directory (will try to mkdir if it doesn\'t exist)')
    args = parser.parse_args()
    print('Query:\n%s' % args.query)
    print('GCS prefix: %s' % args.gcs_prefix)
    if not args.gcs_prefix.startswith('gs://'):
        print('!!! gcs_prefix must start with gs://')
        exit(1)
    local_dir = args.dir
    if local_dir:
        local_dir = os.path.abspath(local_dir)
        print('Local dir: %s' % local_dir)
        if os.path.exists(local_dir):
            if not os.path.isdir(local_dir):
                print('!!! local path exists but it is not a directory')
                exit(1)
        else:
            os.makedirs(local_dir, mode=0o755)

    temp_table = async_query(args.query)
    gcs_pattern = args.gcs_prefix + '-*.jsonl.gz'
    gcs_cleanup(gcs_pattern)
    export_table(temp_table, gcs_pattern)
    if local_dir:
        local_copy(gcs_pattern, local_dir)



import psycopg2
import argparse
import time
import sys
import csv
import boto3
import os
import tempfile
import gzip
from threading import Thread
from queue import Queue
import datetime
import io

# Improve performance by not creating decimal objects - use native floats instead
DEC2FLOAT = psycopg2.extensions.new_type(
                psycopg2.extensions.DECIMAL.values,
                'DEC2FLOAT',
                lambda value, curs: float(value) if value is not None else None)
psycopg2.extensions.register_type(DEC2FLOAT)

# No need to convert strings -> timestamp -> string, just use string direct from input
def parse_date(str_val, cursor):
    if str_val is None:
        return None
    return str_val
    # values = str_val.split("-")
    # return "%d-%02d-%02d" % (values[0], values[1], values[2])
    # return datetime.datetime(int(values[0]), int(values[1]), int(values[2]))

DATE2TIMESTAMP = psycopg2.extensions.new_type(
                psycopg2.extensions.DATE.values,
                'DATE2TIMESTAMP', parse_date)
psycopg2.extensions.register_type(DATE2TIMESTAMP)


parser = argparse.ArgumentParser(description='Clone a redshift db table')
parser.add_argument('--src', type=str, help='host:port:db:user:pass:schema:table', required=True)
parser.add_argument('--dst', type=str, help='host:port:db:user:pass:schema:table', required=True)
parser.add_argument('--s3_key_id', type=str, help='AWS user id for S3', required=True)
parser.add_argument('--s3_secret_key', type=str, help='AWS secret key for S3', required=True)
parser.add_argument('--is_test', type=bool, default = False)
parser.add_argument('--where', type=str, default = None, help = "Where clause limiting data import")
parser.add_argument('--upload_threads', type=int, default = 1, help = "Number of threads to use for uploading data")
parser.add_argument('--client', type=str, required = True, help = "Name of client this table is for")
parser.add_argument('--mod_column', type=str, required = True, help = "Column to mod table by")
parser.add_argument('--columns', type=str, default=None, help ="Comma seperated list of column to copy.  If not set all columns are copied")
parser.add_argument('--from_redshift', type=bool, default = True, help = "Set if src is a redshift table")
parser.add_argument('--src_region', type=str, default ='us-east-1', help = "AWS region where src Redshift is")
parser.add_argument('--distcolumn', type=str, default = None, help = "Dist column for created table")
parser.add_argument('--sortcolumn', type=str, default = None, help = "Sort column for created table")
args = parser.parse_args()
BATCH_SIZE = 50000 if not args.is_test else 10000

if args.src_region == "us-east-1":
    S3_BUCKET = "bg-incoming-east"
else:
    # TODO(jjs): Handle other regions
    S3_BUCKET = "bg-incoming-data"

def get_s3_connection():
    return boto3.client('s3',
            aws_access_key_id=args.s3_key_id,
            aws_secret_access_key=args.s3_secret_key)

def get_s3_dir(client, table, now):
    date = now.strftime("%Y%m%d")
    return "%s/%s_data/%s" % (client, table, date)

class Worker(Thread):
    def __init__(self, table, date, queue, output_files):
        super(Worker, self).__init__()
        self.daemon = True
        self._queue = queue
        self._table = table
        self._date = date
        self._output_files = output_files

    def run(self):
        s3 = get_s3_connection()

        while True:
            chunk_count, rows = self._queue.get()
            with tempfile.NamedTemporaryFile(suffix = ".csv.gz", delete = False) as file:
                gz_file = gzip.GzipFile(fileobj=file, compresslevel=5)
                csv_file = csv.writer(gz_file, quotechar = None, delimiter = '|', quoting = csv.QUOTE_NONE, escapechar='\\')
                for row in rows:
                    output_row = []
                    csv_file.writerow(row)

                gz_file.close()
            with open(file.name, "r") as upload_file:
                key = "%s/%s_%04d.csv.gz" % (get_s3_dir(args.client, self._table, self._date), self._table, chunk_count)
                # print ("file %s bucket: %s key: %s" % (file.name, S3_BUCKET, key))

                s3.put_object(Bucket = S3_BUCKET, Key = key, Body = upload_file)
                self._output_files.append(key)
                self._queue.task_done()
                os.remove(file.name)

def parse_params(params):
    s = params.split(':')
    assert len(s) == 7
    conn = psycopg2.connect(host=s[0], port=s[1], dbname=s[2], user=s[3], password=s[4], sslmode='require')
    conn.autocommit = True
    cur = conn.cursor()
    if s[5]:
        cur.execute("set search_path to '%s'" % s[5])
    return conn, cur, s[6]

def get_select(selected_columns, add_mods):
    where = args.where if args.where else ""
    if add_mods:
        where = where + (" AND " if args.where is not None else "") + "mod(%s,%d) = %%(mods)s" % (args.mod_column, mods)

    if len(where) > 0:
        where = "WHERE %s" % where

    return "select %s from %s %s" % (",".join(selected_columns) if selected_columns is not None else "*", src_table, where)

def run_select_from_redshift(now, selected_columns):
    start_time = time.time()
    src_cur.execute('set query_group TO "slow"')
    statement = get_select(selected_columns, False)
    statement = statement.replace("'", "\\'")
    dest_csv = "%s/%s." % (get_s3_dir(args.client, src_table, now), dst_table)
    manifest = dest_csv + "manifest"
    statement = """
    unload('%s') to 's3://%s/%s'
    null as '__NULL_VALUE__'
    credentials 'aws_access_key_id=AKIAIGHWWWWWV534ALTA;aws_secret_access_key=%s'
    ALLOWOVERWRITE PARALLEL ON GZIP ESCAPE MANIFEST
    """ % (statement, S3_BUCKET, dest_csv, args.s3_secret_key)

    print("statement: %s" % statement)
    src_cur.execute(statement)
    return manifest

def run_select(now, selected_columns):
    s3_files = []
    start_time = time.time()
    print ("Counting rows")
    src_cur.execute('set query_group TO "slow"')
    src_cur.execute("select count(1) from %s %s" % (src_table, ("WHERE %s" % args.where) if args.where else ""))
    row_count = src_cur.fetchone()[0]
    mods = row_count / 100000
    base_statement = get_select(selected_columns, True)
    print("Statement: %s" % base_statement)
    print("Processing %d rows in %d steps" % (row_count, mods))

    record_queue = Queue(args.upload_threads)
    for thread in range(args.upload_threads):
        worker = Worker(dst_table, now, record_queue, s3_files)
        worker.start()

    format_string = '(%s)' % ', '.join(['%s'] * len(col_formats))
    rows = []
    chunks = 0
    row_count = 0
    for mod in range(mods):
        src_cur.execute(base_statement, {'mods': mod})
        for row in src_cur:
            rows.append(row)
            row_count += 1
            if row_count % 10000 == 0:
                sys.stdout.write("Processed %d rows in %ds    \r" % (row_count, time.time() - start_time))
                sys.stdout.flush()

            if row_count % BATCH_SIZE == 0:
                record_queue.put((chunks, rows))
                rows = []
                chunks += 1

    if len(rows) > 0:
        record_queue.put((chunks, rows))
        chunks += 1

    print("")
    while record_queue.qsize() > 0:
        chunks_done = chunks - record_queue.qsize()
        runtime = time.time() - start_time
        rate = max(1, chunks_done * BATCH_SIZE / runtime)
        sys.stdout.write("Waiting for files to upload: %3d ETA: %ds      \r " % (record_queue.qsize(), record_queue.qsize() * BATCH_SIZE / rate))
        sys.stdout.flush()
        time.sleep(2)

    record_queue.join()
    return s3_files

def copy_from_s3(manifest):
    manifest_url = 's3://%s/%s' % (S3_BUCKET, manifest)
    print("Manifest: %s" % manifest_url)
    credentials = 'aws_iam_role=arn:aws:iam::087456767550:role/redshift-data-loader'
    upload_statement = "copy %s from '%s' credentials '%s' MANIFEST ESCAPE GZIP COMPUPDATE ON null as '__NULL_VALUE__' REGION '%s'" % (
            dst_table, manifest_url, credentials, args.src_region)
    upload_start = time.time()
    print("Upload: %s" % upload_statement)
    dst_cur.execute(upload_statement)
    print("Upload time: %5.3fs" % (time.time() - upload_start))

total_time = time.time()
now = datetime.datetime.today()
print ("Connecting to SRC and DEST")
src_conn, src_cur, src_table = parse_params(args.src)
dst_conn, dst_cur, dst_table = parse_params(args.dst)

# Don't think schema changes are transactionable anyway..
# dst_conn.autocommit = True

print ("Creating DST table")
src_cur.execute("select * from pg_table_def where tablename = '%s'" % src_table)
# Ordering is important, so don't use hash/set
if args.columns is not None:
    selected_columns = str.split(args.columns, ',')
    col_formats = [""] * len(selected_columns)
else:
    selected_columns = None
    col_formats = []

for row in src_cur.fetchall():
    name = row[2]
    datatype = row[3]
    col_desc = '%s %s' % (name, datatype)
    if selected_columns is None:
        col_formats.append(col_desc)
    elif name in selected_columns:
        index = selected_columns.index(name)
        if index is not None:
            col_formats[index] = col_desc

if args.from_redshift:
    manifest = run_select_from_redshift(now, selected_columns)
    # manifest = "apartmentlist/lead_scoring_set_with_model_data/20171115/bytegain_leads.manifest"
else:
    s3_files = run_select(now, selected_columns)


dist_text = ("distkey(%s)" % args.distcolumn) if args.distcolumn else ""
sort_text = ("sortkey(%s)" % args.sortcolumn) if args.sortcolumn else ""
dst_cur.execute("drop table if exists %s" % dst_table)
create_str = "create table %s (%s) %s %s" % (dst_table, ', '.join(col_formats), dist_text, sort_text)
print("Executing: %s" % create_str)
dst_cur.execute(create_str)

if args.from_redshift:
    manifest_url = copy_from_s3(manifest)
else:
    print("\nCopying to S3")
    with tempfile.NamedTemporaryFile(suffix = "manifest.json", delete = False) as f:
        f.write('{"entries": [\n')
        text = ',\n'.join(['{"url":"s3://%s/%s","mandatory":true}' % (S3_BUCKET, entry) for entry in s3_files])
        f.write(text)
        f.write("\n]\n}")
        s3 = get_s3_connection()
        manifest_key = "%s/%s.manifest" % (get_s3_dir(args.client, src_table, now), dst_table)
        f.flush()
        f.seek(0)
        s3.put_object(Bucket = S3_BUCKET, Key = manifest_key, Body = f)

    manifest_url = 's3://%s/%s' % (S3_BUCKET, manifest_key)
    credentials = 'aws_iam_role=arn:aws:iam::087456767550:role/redshift-data-loader'
    upload_statement = "copy %s from '%s' credentials '%s' MANIFEST ESCAPE GZIP COMPUPDATE ON EMPTYASNULL;" % (dst_table, manifest_url, credentials)
    upload_start = time.time()
    dst_cur.execute(upload_statement)
    print("Upload time: %5.3fs" % (time.time() - upload_start))

print("Total clone time: %ds" % (time.time() - total_time))



import pickle

from bytegain.data_import.segment_utils import parse_s3_url

# file = open("bytegain/cust/apartmentlist/data/all_features/lead_scoring_lead_scoring_set/lead_scoring.lead_scoring_set..pickle")
filespec = "gs://bg-dataflow/all_features"
# filespec = "gs://bg-dataflow/all_features/web_pages/"

filespec = "gs://bg-dataflow/apartmentlist/user_value/1year_date_fix"


def get_schema(filespec, subdirs={'events':'/web_pages/', 'features':'/lead_scoring_lead_scoring_set_with_model/',
                                  'identifies':'/api_identifies/', 'interest':'/api_interest/'}):
    from google.cloud import storage
    client = storage.Client()
    bucket_name, key, _ = parse_s3_url(filespec)
    bucket = client.get_bucket(bucket_name)
    tables_info = {}
    for k, v in list(subdirs.items()):
        path = key + v
        blobs = bucket.list_blobs(prefix=path)
        for blob in blobs:
            if (blob.name).endswith('.pickle'):
                file = blob
        pickle_string = file.download_as_string()
        table_info = pickle.loads(pickle_string)
        tables_info[k] = table_info
    return tables_info

schemas = get_schema(filespec)
features = schemas['features']
identifies = schemas['identifies']

sample = {"budget_min":"630","dwell_time":"149","commute_minutes":"","user_id":"236156686",
          "interest_state":"strong_interest","viewed_availability":"t","registered_source_partner":"apartment_list",
          "interest_source_app":"MobileApartmentList","property_lat":"39.974504",
          "preferred_budget_deviation":"1.04046242","preferred_available_units":"45","label":"f",
          "positive_velocity_rate":"0.53508771929824561","preferred_units":"75","search_category":"Email",
          "registered_source_app":"MobileApartmentList","budget_max":"900",
          "preferences_move_urgency":"0.80000000000000004","timestamp":"2017-01-11 06:00:14.914","move_in_delay":"-123",
          "metro_id":"747240","preferences_lease_length":"12","commute_distance":"","commute_mode":"",
          "preferences_baths":"","property_lon":"-86.080742","distance":"2.861052221994","rental_id":"p10663",
          "interest_source_partner":"apartment_list","median_preferred_price":"865.0","preferences_beds_2":"f",
          "preferences_beds_3":"f","preferences_beds_0":"f","preferences_beds_1":"t","property_zip_code":"46033",
          "email_clicked":"t","average_preferred_price":"904","has_phone_number":"t","data_updated_at":"2017-10-18 23:58:33"}


# Maps categorical feature to index
def category_to_index(groups):
    cat_ind_map = {}
    for index, (name, fraction) in enumerate(groups):
        if not name:
            name = ''
        cat_ind_map[name] = index
    if '' not in cat_ind_map:
        cat_ind_map[''] = index + 1
    return cat_ind_map

def numerical_to_index(value, vals, num_buckets):
    import numpy as np
    (median, perc25, perc75) = vals
    bins = np.arange(perc25, perc75, abs(perc75 - perc25) / num_buckets)
    bins = np.insert(bins, 0, -np.inf)
    bins = np.append(bins, np.inf)
    if value is None:
        return len(bins), len(bins) + 1
    # TODO: Add minus and plus infinities into bins
    index = np.digitize(value, bins) - 1
    return index, len(bins) + 1

from bytegain.custom.cust.apartmentlist.leads_filter_160 import EXCLUDED_FEATURES as excluded_features
# Maps each feature to index
def get_maps(features):
    maps = {}
    cats, bools, other, time, excluded = 0, 0, 0, 0, 0
    excluded_set = set(excluded_features)
    for col in features.columns:
        if col.name in excluded_set:
            excluded += 1
            continue
        if 'character' in col.datatype:
            if col.name == 'rental_id':
                continue
            maps[col.name] = category_to_index(sorted(col._groups, key=lambda x: x[1]))
            cats += 1
        elif col.datatype == 'boolean':
            maps[col.name] = {'t': 1, 'f': 0}
            bools += 1
        elif 'time' in col.datatype:
            time += 1
        else:
            vals = col.median, col.percentile_25, col.percentile_75
            maps[col.name] = lambda x, vals=vals: numerical_to_index(x, vals, 7)
            other += 1
    print(("Total features without excluded and time: %d, categorical: %d, boolean :%d, other %d" % (
    len(features.columns) - excluded - time, cats, bools, other)))
    return maps

# def sample_to_one_hot(sample):
#     sample_one_hot = {}
#     for k, v in sample.iteritems():
#         bucket_map = maps.get(k, [])
#         bucket_len = len(bucket_map)
#         if bucket_len > 0:
#             sample_one_hot[k] = [bucket_len, bucket_map.get(v)]
#     return sample_one_hot

def sample_to_vec(sample, maps):
    sample_vec = []
    bucketed_features = 0
    for k, v in list(sample.items()):
        bucket_map = maps.get(k, [])
        if callable(bucket_map):
            num = None if v == '' else float(v)
            index, bucket_len = bucket_map(num)
        else:
            bucket_len = len(bucket_map)
            index = None
        print('Feature: %s' % k)
        if bucket_len > 0:
            vec = [0] * bucket_len
            if index is not None:
                vec[index] = 1
            else:
                vec[bucket_map.get(v)] = 1
            sample_vec += vec
            bucketed_features += 1
    print('Sample: total features %s, bucketed features %s' % (len(sample), bucketed_features))
    return sample_vec

# import time
# time0 = time.time()
# maps = get_maps(features)
# # for k, v in maps.iteritems():
# #     print '%s: %s'%(k, v)
# maps_time = time.time()
# # sample_one_hot = sample_to_one_hot(sample)
# # print 'Bucketed sample: %s' % sample_one_hot
# vec = sample_to_vec(sample, maps)
# vec_time = time.time()
# print 'Len vec %d, vec: %s' % (len(vec), vec)
# print 'Maps time %.4f' % (maps_time - time0)
# print 'Vec time %.4f' % (vec_time - maps_time)

from bytegain.custom.model_data.bucket_features import FeatureBucketerLinear as feature_bucketer
bucketer = feature_bucketer(excluded_features, features)
bucketer.construct_buckets()
vec_bucketed = bucketer.sample_to_vec(sample)
print('Length %d, vec: %s' % (len(vec_bucketed), vec_bucketed))

print('Nonlinear bucketer')
from bytegain.custom.model_data.bucket_features import FeatureBucketerNonLinear as feature_bucketer
bucketer = feature_bucketer(excluded_features, features)
bucketer.construct_buckets()
vec_bucketed = bucketer.sample_to_vec(sample)
print('Length %d, vec: %s' % (len(vec_bucketed), vec_bucketed))

print('Direct bucketer')
from bytegain.custom.model_data.bucket_features import FeatureBucketerDirect as feature_bucketer
bucketer = feature_bucketer(excluded_features, features)
bucketer.construct_buckets()
vec_bucketed = bucketer.sample_to_vec(sample)
print('Length %d, vec: %s' % (len(vec_bucketed), vec_bucketed))



"""
Bootstrap AUC confidence intervals.

Example:

python calc.py --input al_data_both_smooth.csv --group model --prediction score --label leased --iterations 10

"""
import argparse
import csv
import sklearn.metrics
import numpy as np

from collections import defaultdict

parser = argparse.ArgumentParser(description='Bootstrap AUC confidence intervals')
parser.add_argument('--input', type=str, required=True, help='Input filename')
parser.add_argument('--group', type=str, required=True, nargs='+', help='Columns to concatenate to construct group name')
parser.add_argument('--prediction', type=str, required=True, help='Column name of prediction')
parser.add_argument('--label', type=str, required=True, help='Column name of label')
parser.add_argument('--iterations', type=int, default=1000, help='Number of bootstrap iterations')
parser.add_argument('--filter_groups', type=str, nargs='+', help='Filter to id set available for all specified groups')
parser.add_argument('--filter_id', type=str, nargs='+', help='Columns to concatenate to construct id filtering.')
args = parser.parse_args()

group_predictions = defaultdict(list)
group_labels = defaultdict(list)
group_ids = defaultdict(list)

with open(args.input) as f:
    reader = csv.reader(f)
    header = next(reader)

    group_index_list = [header.index(i) for i in args.group]
    prediction_index = header.index(args.prediction)
    label_index = header.index(args.label)
    id_index_list = [header.index(i) for i in args.filter_id] if args.filter_id else None

    predictions = []
    labels = []
    for row in reader:
        group = ':'.join([row[i] for i in group_index_list])
        if row[prediction_index] == '':
            # TODO: What is this?
            continue
        group_predictions[group].append(float(row[prediction_index]))
        group_labels[group].append(int(row[label_index]))
        if id_index_list:
            id_str = ':'.join([row[i] for i in id_index_list])
            group_ids[group].append(id_str)

if args.filter_groups:
    print("Filtering to ids shared by all of %s..." % ' '.join(args.filter_groups))
    shared_ids = set(group_ids[args.filter_groups[0]])
    for group in args.filter_groups[1:]:
        shared_ids.intersection_update(group_ids[group])
    print("Filtered to %d ids." % len(shared_ids))
    for group in sorted(group_predictions.keys()):
        indexes = np.array([i for i in range(len(group_ids[group])) if group_ids[group][i] in shared_ids])
        group_predictions[group] = [group_predictions[group][x] for x in indexes]
        group_labels[group] = [group_labels[group][x] for x in indexes]
        group_ids[group] = [group_ids[group][x] for x in indexes]

print("Bootstrapping 90%% intervals with %d iterations..." % args.iterations)
for name in sorted(group_predictions.keys()):
    labels = np.array(group_labels[name])
    predictions = np.array(group_predictions[name])

    if args.filter_groups and name not in args.filter_groups:
        continue

    rate_list = []
    auc_list = []
    logloss_list = []
    for iteration in range(args.iterations):
        # Re-sample with replacement
        which = np.random.choice(len(predictions), len(predictions))
        rate_list.append(np.mean(labels[which]))
        try:
            auc_list.append(sklearn.metrics.roc_auc_score(labels[which], predictions[which]))
        except ValueError:
            auc_list.append(0.5)

        pred = predictions[which]
        logloss_list.append(sklearn.metrics.log_loss(labels[which], np.array([np.subtract(1, pred), pred]).T))

    rate_list = sorted(rate_list)
    auc_list = sorted(auc_list)
    logloss_list = sorted(logloss_list)
    print(("%s samples:%6d rate:%.3f [%.3f .. %.3f] auc:%.3f [%.3f .. %.3f] logloss:%.3f [%.3f .. %.3f]" % (
                    name,
                    len(predictions),
                    np.mean(rate_list),
                    rate_list[args.iterations * 5 / 100],
                    rate_list[args.iterations * 94 / 100],
                    np.mean(auc_list),
                    auc_list[args.iterations * 5 / 100],
                    auc_list[args.iterations * 94 / 100],
                    np.mean(logloss_list),
                    logloss_list[args.iterations * 5 / 100],
                    logloss_list[args.iterations * 94 / 100])))



import bytegain.custom.model_data.row_handler_proto.row_handler_pb2 as row_handler_pb2
import bytegain.custom.model_data.bucket as bucket_module

"""
MultiRowHandler converts a series of input rows with raw data into a set of examples for training.
Multiple rows of input can result in a single row of output.
"""

class MultiRowHandler(object):
    def __init__(self, label_bucket = None, id_extractor = None):
        self._label_bucket = label_bucket
        self._id_extractor = id_extractor
        self._buckets = []
        self._width = None
        self._last_id = None
        self._current_events = None

        # if load_from is not None:
        #       self._load_from_file(load_from)

    @property
    def buckets(self):
        return self._buckets

    @property
    def width(self):
        if self._width is None:
            self._width = 0
            for bucket in self._buckets:
                self._width += bucket._total_buckets
        return self._width

    def replace_field_names_with_index(self, cursor):
        # Instead of using a DictCursor, use a regular cursor,
        # and rewrite the extractor field names with array
        # offsets.  This speeds up data loading by about 20%.
        header = [x.name for x in cursor.description]
        for bucket in self._buckets:
            extractor = bucket._extractor
            extractor.update_fields_by_id(header)

        self._id_extractor.update_fields_by_id(header)
        self._label_bucket._extractor.update_fields_by_id(header)

    def _load_from_file(self, filename):
        with open(filename, "rb") as f:
            handler = row_handler_pb2.RowHandler()
            handler.ParseFromString(f.read())

        for bucket in handler.bucket:
            self._buckets.append(bucket_module.read_from_proto(bucket))

    def save(self, filename):
        handler = row_handler_pb2.RowHandler()
        for bucket in self._buckets:
            bucket_proto = handler.bucket.add()
            bucket.write_to_proto(bucket_proto)

        with open(filename, "wb") as f:
            f.write(handler.SerializeToString())

    def add_bucket(self, bucket):
        self._buckets.append(bucket)

    def get_id_from_row(self, row):
        if self._id_extractor == None:
            raise Exception("No ID function given")
        return self._id_extractor.extract(row)

    def get_label_from_row(self, row):
        if self._label_bucket == None:
            raise Exception("No Label function given")
        return self._label_bucket.bucketize(row)

    def _make_entry(self):
        return self._make_data_entry(), self.get_label_from_row(self._current_events[0]), self.get_id_from_row(self._current_events[0])

    def process_row(self, row):
        new_val = None
        row_id = self.get_id_from_row(row)
        if row_id != self._last_id:
            if self._last_id:
                new_val = self._make_entry()
            self._last_id = row_id
            self._current_events = []

        self._current_events.append(row)
        return new_val

    def get_fields(self):
        fields = []
        for bucket in self._buckets:
            fields.extend(bucket._extractor.get_fields())
        return fields



"""
Example usage:
net, results = cross_validate.run(dataset, net_function)
"""

import bytegain.nn_tools.auc_plot as auc_plot

def run(dataset, create_net, batches, runs = 5):
    auc_sum = 0
    aucs = []
    test_results = []
    for run in range(runs):
        net = create_net(dataset);
        aucs.append(net.train(batches, 4000))
        test_results.extend(net.get_sorted_test_labels());
        if run < runs - 1:
            dataset.repartition()
    auc_plot.sort_results(test_results)
    auc = auc_plot.compute_auc(test_results)
    print("Total Auc: %5.3f %s" % (auc, str([("%5.3f" % a) for a in aucs])))

    return net, test_results



import argparse
import os
import sys
import json
import time
from datetime import datetime
from bytegain.iox import gcp_init

# Use cred file and private key to set google credentials
gcp_cred_file = os.environ.get("GCP_CRED_FILE")
gcp_private_key = os.environ.get("GCP_PRIVATE_KEY")
if gcp_cred_file and gcp_private_key:
    gcp_init.set_credentials_file(gcp_cred_file, gcp_private_key)

import googleapiclient.discovery

from bytegain.data_import.segment_utils import parse_s3_url
import configparser

PROJECT_ID = 'tough-hologram-156002'


def create_transfer_client():
    return googleapiclient.discovery.build('storagetransfer', 'v1')


parser = argparse.ArgumentParser()
parser.add_argument('--s3_url', required=True)
parser.add_argument('--gs_url', required=True)
args = parser.parse_args()

s3_bucket, s3_key, _ = parse_s3_url(args.s3_url)
gs_bucket, gs_key, _ = parse_s3_url(args.gs_url)

if "AWS_ACCESS_KEY_ID" in os.environ and "AWS_SECRET_ACCESS_KEY" in os.environ:
    aws_access_key = os.environ["AWS_ACCESS_KEY_ID"]
    aws_secret_key = os.environ["AWS_SECRET_ACCESS_KEY"]
else:
    try:
        config = configparser.ConfigParser()
        config.read([str(os.path.expanduser('~') + "/.aws/credentials")])
        aws_access_key = config.get('default', 'aws_access_key_id')
        aws_secret_key = config.get('default', 'aws_secret_access_key')
    except:
        aws_access_key = None
        aws_secret_key = None

if aws_access_key is None or aws_secret_key is None:
    print("Error: AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must be set either in the environment or in ~/.aws/credentials")
    sys.exit(1)
else:
    print("AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY are set")

now = datetime.now()
# Edit this template with desired parameters.
# Specify times below using US Pacific Time Zone.
transfer_job = {
    'description': "Transfer from %s to %s" % (args.s3_url, args.gs_url),
    'status': 'ENABLED',
    'projectId': PROJECT_ID,
    'schedule': {
        'scheduleStartDate': {
            'day': now.day,
            'month': now.month,
            'year': now.year
        },
        'scheduleEndDate': {
            'day': now.day,
            'month': now.month,
            'year': now.year
        },
    },
    'transferSpec': {
        'transferOptions': {
            'overwriteObjectsAlreadyExistingInSink': True
        },
        'objectConditions': {
            'includePrefixes': [s3_key]
        },
        'awsS3DataSource': {
            'bucketName': s3_bucket,
            'awsAccessKey': {
                'accessKeyId': aws_access_key,
                'secretAccessKey': aws_secret_key
            }
        },
        'gcsDataSink': {
            'bucketName': gs_bucket
        }
    }
}

client = create_transfer_client()
result = client.transferJobs().create(body=transfer_job).execute()
# print('Returned transferJob: {}'.format(
#     json.dumps(result, indent=4)))
job_name = result["name"]
op_filter = {
    'project_id': PROJECT_ID,
    'job_names': [job_name]
}
start_time = time.time()
# Wait for operation to be created
time.sleep(30)

print("Waiting for %s" % job_name)
while True:
    ops = client.transferOperations().list(name='transferOperations', filter=json.dumps(op_filter)).execute()
    if "operations" in ops:
        ops = ops["operations"]
        if len(ops) != 1:
            print("Bad ops returned: %s" % ops)
            break
        else:
            op = ops[0]
            # print json.dumps(op, indent=4)
            if "done" in op and op["done"]:
                total_time = time.time() - start_time
                data = int(op["metadata"]["counters"]["bytesCopiedToSink"])
                print(("Files Copied %5.3fG in %ds (%4.1fM/s)" % (
                    data / (1024.0 * 1024 * 1024), total_time, data / (1024 * 1024.0) / total_time)))
                break
            elif op["metadata"]["status"] == "IN_PROGRESS":
                total_time = time.time() - start_time
                if "bytesCopiedToSink" in op["metadata"]["counters"]:
                    data = int(op["metadata"]["counters"]["bytesCopiedToSink"])
                    sys.stdout.write("Data: %5.3fG in %ds (%4.1fM/s)    \r" % (
                        data / (1024.0 * 1024 * 1024), total_time, data / (1024 * 1024.0) / total_time))
                    sys.stdout.flush()

    time.sleep(10)



import argparse
import logging
import apache_beam as beam
from bytegain.beam.beam_util import encode_json, DoFnWithCounters, run_pipeline, create_pipeline, get_counter
from bytegain.zap.event_utils import decode_json, _is_properties_broken
from datetime import datetime, timedelta
import math

logger = logging.getLogger('events_to_sessions')


def quantize(value, max_linear=100):
    if value > max_linear:
        normed = value * 1.0 / max_linear
        pow = int(math.log(normed, 2)) + 1
        value = int(max_linear * (2 ** pow))
    return value



class MergeEvents(DoFnWithCounters):
    """Create sessions for only users who signed up. Merge leases from _acappella_leases and events from BQ"""
    def __init__(self, only_signed, downsample_rate=1):
        super(MergeEvents, self).__init__()
        self._only_signed = only_signed
        self._downsample_rate = downsample_rate

    def process(self, xxx_todo_changeme):
        (user, data) = xxx_todo_changeme
        from bytegain.beam.beam_util import get_counter
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch, get_timestamp

        (_acappella_leases, _events) = data

        if user is None:
            get_counter('merge_events', 'user_id_none').inc()
            return

        # Take first 5000 events (_events is iterable), convert it to list here
        first_events = []
        for event in _events:
            if len(first_events) > 5000:
                break
            first_events.append(event)
        _events = first_events

        # Use only the first lease (one lease per user)
        first_lease = []
        for lease in _acappella_leases:
            first_lease.append(lease)
            break
        _acappella_leases = first_lease
        if _acappella_leases:
            get_counter('merge_events', 'users_with_lease').inc()

        if len(_events) > 0:
            get_counter('merge_events', 'users_with_events').inc()
            get_counter('merge_events', 'session_len_%04d' % quantize(len(_events))).inc()

            if _acappella_leases:
                get_counter('merge_events', 'users_with_events_and_lease').inc()
            lease = len(_acappella_leases) > 0

            # Find session duration and whether the user took a quiz
            signup = False
            earliest_timestamp = latest_timestamp = timestamp_to_epoch(get_timestamp(_events[0]))
            for event in _events:
                signup = event.get('context', {}).get('page', {}).get('search') == '?signup=true' or signup
                timestamp_event = timestamp_to_epoch(get_timestamp(event))
                if timestamp_event < earliest_timestamp:
                    earliest_timestamp = timestamp_event
                if timestamp_event > latest_timestamp:
                    latest_timestamp = timestamp_event

            # Counters
            if signup:
                get_counter('merge_events', 'users_with_quiz').inc()
                if _acappella_leases:
                    get_counter('merge_events', 'users_with_quiz_and_lease').inc()
            else:
                get_counter('merge_events', 'users_without_quiz').inc()

            session_duration = latest_timestamp - earliest_timestamp
            if session_duration < 60:
                get_counter('merge_events', 'session_duration_mins_00_lease_%s' % lease).inc()
            elif session_duration < 3600:
                get_counter('merge_events',
                            'session_duration_mins_%02d_lease_%s' % (session_duration / 60, lease)).inc()
            elif session_duration < 3600 * 24:
                get_counter('merge_events',
                            'session_duration_hrs_%02d_lease_%s' % (session_duration / 3600, lease)).inc()
            else:
                get_counter('merge_events', 'session_duration_days_%02d_lease_%s' % (
                session_duration / 3600 / 24, lease)).inc()
            # Downsample sessions (WARNING: this breaks p_vs_a! Prediction is not a true likelihood anymore)
            if self._downsample_rate != 1:
                if not lease and (hash(_events[-1]['timestamp']) % self._downsample_rate) != 0:
                    return
            # Use only users who:
            # 1) signed-up if `_only_signed` is set
            # 2) or all users
            if (self._only_signed and signup) or not self._only_signed:
                # Sort events by timestamp
                _events = sorted(_events + _acappella_leases, key=lambda _event: _event['timestamp'])
                yield user, _events
        else:
            get_counter('merge_events', 'users_without_events').inc()


def make_sessions(driver_input, session_generator_config, dataflow_args, output_sessions, tmp_dir):
    """
    Called by the zap driver, from the sessions stage.
    """
    p = argparse.ArgumentParser()
    args, _ = p.parse_known_args(driver_input.extra_args)

    bq_dataset = session_generator_config.get('bq_dataset')
    assert bq_dataset, 'bq_dataset must be set by the driver'

    num_shards = session_generator_config['num_shards']
    assert num_shards >= 0

    data_start_date = session_generator_config['data_start_date']
    data_end_date = session_generator_config['data_end_date']

    p = create_pipeline(dataflow_args, 'al-bg-ltv-e2s')

    t = 'acappella_leases'
    table = '`bg-event-ingester.%s.%s`' % (bq_dataset, t)
    where = " timestamp BETWEEN TIMESTAMP('{data_start_date}') AND TIMESTAMP('{data_end_date}')".format(
        data_start_date=data_start_date, data_end_date=data_end_date)
    bq_query_leases = "SELECT *, 'acappella_leases' as type FROM {table} WHERE {where}".format(table=table, where=where)

    logger.info('Leases query: %s', bq_query_leases)

    acappella_leases = (p
                        | 'read %s' %t >> beam.io.Read(
                beam.io.BigQuerySource(query=bq_query_leases, validate=True, use_standard_sql=True))
                        | 'map to userId in %s' % t >> beam.Map(lambda x: (str(x['userId']).decode('utf-8'), x))
                        )

    # Overwrite BQ query
    limit_users = session_generator_config.get('limit_users', 0)

    # For sessions don't use last X days
    end_date = datetime.strptime(data_end_date, '%Y-%m-%d')
    blackout_days = session_generator_config['blackout']
    end_date = end_date - timedelta(days=int(blackout_days))
    data_end_date_sessions = end_date.strftime('%Y-%m-%d')

    t = 'events'
    table = '`bg-event-ingester.%s.%s`' % (bq_dataset, t)
    where = "IFNULL(userId, '') <> '' AND timestamp BETWEEN TIMESTAMP('{data_start_date}') AND TIMESTAMP('{data_end_date}')".format(
        data_start_date=data_start_date, data_end_date=data_end_date_sessions)
    where += " AND (event NOT IN ('BG Scroll', 'BG Click', 'BG Click2', 'BG Key', 'bytegain_predict'))"
    if limit_users > 0:
        where += ' AND anonymousId IN (SELECT DISTINCT anonymousId FROM {table} WHERE {where} ORDER BY 1 LIMIT {limit})'.format(
            table=table, where=where, limit=limit_users)
    bq_query_events = 'SELECT * FROM {table} WHERE {where}'.format(table=table, where=where)

    logger.info('Events query: %s', bq_query_events)

    bq_events = (p
                 | 'read %s' % t >> beam.io.Read(
        beam.io.BigQuerySource(query=bq_query_events, validate=True, use_standard_sql=True))
                 | 'extract json' >> beam.Map(lambda x: x['json'])
                 | 'decode json' >> beam.Map(decode_json)
                 | 'filter bad properties' >> beam.Filter(_is_properties_broken)
                 | 'map to userId' >> beam.Map(lambda x: (x['userId'], x))
                 )

    # Merge acappella leases and bg sessions
    ((acappella_leases, bq_events) | 'group acappella leases and bg events' >> beam.CoGroupByKey()
     | 'merge and sort events' >> beam.ParDo(MergeEvents(session_generator_config['only_signed'],
                                                         downsample_rate=session_generator_config.get('downsample', 1)))
     | 'encode json' >> beam.Map(encode_json, counter=get_counter("make_sessions", "nonempty_sessions"))
     | 'save' >> beam.io.WriteToText(output_sessions, file_name_suffix='.gz', num_shards=num_shards)
     )
    counters = run_pipeline(p)

    return {'counters': counters,
            'data_end_date_sessions': data_end_date_sessions,
            'bq_query_events': bq_query_events,
            'bq_query_leases': bq_query_leases}



import argparse
import csv
import http.client
import json
import importlib
from collections import defaultdict

import psycopg2
import psycopg2.extras

import bytegain.custom.cust.apartmentlist.al_xgb as al_xgb
# import bytegain.custom.cust.apartmentlist.search_data as search_data
import bytegain.custom.model_data.read_from_db as read_from_db
import bytegain.custom.nn_tools.model_utils as model_utils
from bytegain.custom.model_data.extractor import RowExtractor
from bytegain.custom.nn_tools.prod_classifier import ProdXgbClassifier

parser = argparse.ArgumentParser(description='Convert AL data to json')
parser.add_argument('--count', type=int, default=100, help='# examples to verify')
parser.add_argument('--server', type=str, default="localhost:5000")
parser.add_argument('--redshift_url', type=str, default=None)
parser.add_argument('--api_key', type=str, default="AL_TESTING")
parser.add_argument('--model', type=str, help="name of model to test")
parser.add_argument('--model_dir', type=str, help="model directory- if not set load from S3", default=None)
parser.add_argument('--model_config', type=str, help = "python var with model config")
parser.add_argument('--random_seed', type=int, default=9232323)
parser.add_argument('--epsilon', type=float, default=1E-5)
parser.add_argument('--local_model', type=bool, default=False, help="If set, then load model locally for testing")
parser.add_argument('--save_json', type=str, default=None,
                    help="If set, then save requests as json files instead of sending to server")

args = parser.parse_args()

user_ids = set()


def get_config(model_config):
    module, var = model_config.rsplit('.', 1)
    module = importlib.import_module(module)
    return getattr(module, var)

def get_input_rows(config, fields, sortby="random()"):
    model_name = "%s_%s" % (config['model_name'], config['model_version'])
    if "search" in model_name:
        pass
        # selector = search_data.get_selector()
        # extractor = search_data.SearchIdExtractor()
    else:
        selector = al_xgb.get_selector(config)
        extractor = RowExtractor(config['id_field'])

    statement = "select * from %s %s order by %s limit %d" % (
        selector._table, ("where %s" % selector._where) if selector._where is not None else "", sortby,
        args.count * 100)

    with read_from_db.create_connection() as connection:
        all_rows = []
        fields.extend(extractor.get_fields())
        with connection.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
            cursor.execute("set seed to %f" % (1.0 / args.random_seed))
            cursor.execute(statement)
            for row in cursor:
                dict_row = {}
                for field in fields:
                    dict_row[field] = row[field]
                    if field == 'message_counts':
                        dict_row[field] = None

                all_rows.append(dict_row)

    unique_users = set()
    rows = {}
    for row in all_rows:
        id = extractor.extract(row)
        if id not in unique_users:
            if len(unique_users) == args.count * 10:
                break
            unique_users.add(id)
        rows[str(id)] = row
    return rows



def load_model(model_name):
    model_spec = model_utils.XgbModelSpec(model_name, args.model_dir)

    if args.model_dir is None:
        model_spec.download(for_prod=False)

    return model_spec, ProdXgbClassifier(model_spec)


def get_local_prediction(classifier, row):
    return classifier.classify(row)


def get_json(row):
    container = {'api_key': args.api_key, 'services': [{'service': 'model', 'models': model_name, 'data': row}]}
    return json.dumps(container, indent=4, sort_keys=True)


def get_server_prediction(connection, row):
    json_data = get_json(row)
    # print ("Post: %s" % str(json_data))
    connection.request("POST", "/apartmentlist/lease_predict", json_data)
    response = connection.getresponse()
    body = response.read()
    # print ("Response: %s" % str(json_response))
    try:
        json_response = json.loads(body)
        prediction = json_response["responses"][0]["predictions"][0]["raw_score"]
    except Exception as e:
        print("Post: %s" % str(json_data))
        print("Response: %s" % str(body))
        raise
    return prediction


def get_server_search_prediction(connection, json_data):
    # print ("Post: %s" % str(json_data))
    connection.request("POST", "/apartmentlist/search_rank", json_data)
    response = connection.getresponse()
    body = response.read()
    json_response = json.loads(body)
    # print ("Response: %s\n\n" % str(json_response))
    try:
        prediction = json_response["responses"][0]
    except Exception as e:
        print("Post: %s" % str(json_data))
        print("Response: %s" % str(json_response))
        raise
    return prediction


def get_train_predictions(ids, csv_file):
    predictions = {}
    ids = set(ids)
    with open(csv_file, "r") as f:
        reader = csv.reader(f)
        # Skip header
        next(reader)
        for line in reader:
            id = line[0]
            if id in ids:
                predictions[id] = float(line[1])

    print("found %d of %d" % (len(predictions), len(ids)))
    return predictions


def group_by_id(rows):
    by_id = defaultdict(lambda: [])
    for k, v in list(rows.items()):
        id = k.split("_")[0]
        by_id[id].append(v)

    return by_id


def find_row_by_order_id(train_rows, id):
    for row in train_rows:
        if int(row['id'].split("_")[2]) == id:
            return row
    return None


def process_search(db_rows, train_rows, row_handler):
    new_train_rows = {}
    for k, v in list(train_rows.items()):
        # print v
        new_train_rows[k] = {'prob': v, 'id': k}

    db_rows = group_by_id(db_rows)
    train_rows = group_by_id(new_train_rows)
    count = 0
    miss_count = 0
    rh_keys = set(row_handler.get_fields())
    rh_keys.add('unit_id')
    for id, user_db_rows in list(db_rows.items()):
        # print "Id: %s" % id
        if count == args.count:
            break
        if id in train_rows and len(user_db_rows) > 5:
            count += 1
            process_unit_rows(rh_keys, user_db_rows)
            user_train_rows = train_rows[id]
            existing_values = user_db_rows[0]
            user_keys = set(existing_values.keys())
            for row in user_db_rows:
                for k, v in list(row.items()):
                    if k in user_keys and existing_values[k] != v:
                        user_keys.remove(k)
            user_values = {}

            for k in user_keys:
                user_values[k] = existing_values[k]
                for row in user_db_rows:
                    del row[k]

            container = {'api_key': args.api_key,
                         'services': [{'service': 'search',
                                       'model': model_name,
                                       'user': user_values,
                                       'units': user_db_rows}]}

            json_data = json.dumps(container, indent=4, sort_keys=True)
            # print json_data

            prediction = get_server_search_prediction(connection, json_data)
            if len(prediction) != len(user_db_rows):
                print("Wrong number of predictions: %d != %d" % (len(prediction), len(user_db_rows)))
                print(prediction)
            miss = False
            for result in prediction:
                id = result['unit_id']
                raw_score = result['raw_score']
                train_row = find_row_by_order_id(user_train_rows, id)
                if train_row == None:
                    print("could not find row: %d in %s" % (id, user_train_rows))
                    miss = True
                else:
                    # print train_row
                    if abs(train_row['prob'] - raw_score) > 0.00001:
                        miss = True
                        print("Missed %s %f != %f" % (id, raw_score, train_row['prob']))
            if miss:
                miss_count += 1
    print("Missed: %d of %d" % (miss_count, count))


def process_unit_rows(rh_keys, user_db_rows):
    for row in user_db_rows:
        row['unit_id'] = row['user_order']
    for row in user_db_rows:
        keys = set(row.keys())
        keys.difference_update(rh_keys)
        for k in keys:
            del row[k]


def process_lease(input_rows, train_rows, row_handler):
    count = 0
    diff = 0
    misses = []
    pred_sum = 0.0
    for id in input_rows:
        if count == args.count:
            break
        if id in train_rows:
            count += 1
            row = input_rows[id]
            if args.save_json:
                print ("Saving json")
                json_data = get_json(row)
                with open(args.save_json % count, "wt") as f:
                    f.write(json_data + "\n")
            else:
                if connection:
                    prediction = get_server_prediction(connection, row)
                else:
                    prediction = get_local_prediction(classifier, row)

                train_pred = train_rows[id]
                pred_sum += prediction
                delta = train_pred - prediction
                if abs(delta) > args.epsilon:
                    # print ("Delta: %20s %6.4f %6.4f : %7.5f %s" % (id, prediction, train_pred, delta, str(row)))
                    # if classifier is not None:
                    #     print ("input row: %s" % str(classifier._row_handler._row_handler.get_data_from_row(row)))
                    # print "--------------"
                    # misses.append(row)
                    print("%5.3f -> %5.3f (%3d%%)" % (train_pred, prediction, prediction * 100/train_pred - 100))
                    diff += 1
    print("%5.3f%% same" % (100 - (diff * 100.0 / count)))
    print("Avg prediction: %6.4f" % (pred_sum /count))
    if count != args.count:
        print("Too few examples: %d" % count)
    if misses > 1:
        miss_values = defaultdict(lambda: defaultdict(lambda: 0))
        for row in misses:
            for k, v in list(row.items()):
                miss_values[k][v] = miss_values[k][v] + 1
        for k, v in list(miss_values.items()):
            print("%s: %s" % (k, str(["%s=%s" % (x, y) for x, y in list(v.items())])))

config = get_config(args.model_config)
model_name = "%s_%s" % (config['model_name'], config['model_version'])

model_spec, classifier = load_model(model_name)
# classifier._row_handler.print_summary()
# exit(0)
# model_name = 'al_finance_1.0.0'
row_handler = classifier._row_handler
input_rows = get_input_rows(config, classifier.get_fields(), "md5(user_id)" if "search" in model_name else "random()")

if args.local_model:
    print ("Testing against local model")
    row_handler.print_spec()
    connection = None
elif args.save_json is None:
    print("Testing against %s" % (args.server))
    classifier = None
    connection = http.client.HTTPSConnection(args.server)

train_rows = get_train_predictions(list(input_rows.keys()), model_spec.get_validation_file())

if "search" in model_name:
    process_search(input_rows, train_rows, row_handler)
else:
    process_lease(input_rows, train_rows, row_handler)



import argparse
from bytegain.custom.data_import.schema_info import TableInfo, SchemaInfo, read_schema_from_file, save_schema
import math
import json

class SchemaFilter(object):
    def __init__(self):
        self.max_null = 0.95
        self.max_len = 40
        self.min_total_category_fraction = 0.10
        self.min_category_fraction = 0.01
        self.min_join_table_row_fraction = 0.01
        self.tables_to_keep = []
        self.tables_to_remove = []
        self.columns_to_keep = []
        self.ignored_events = []

    @staticmethod
    def load_from_json(json_file):
        with open(json_file, "r") as f:
            content = json.load(f)

        filter = SchemaFilter()
        filter.max_null = content['max_column_null']
        filter.max_len = content['max_avg_len']
        filter.min_total_category_fraction = content['min_total_category_fraction']
        filter.min_category_fraction = content['min_category_fraction']
        filter.min_join_table_row_fraction = content['min_join_table_row_fraction']
        filter.tables_to_keep = content['tables_to_keep']
        filter.tables_to_remove = content['tables_to_remove']
        filter.columns_to_keep = content['columns_to_keep']
        filter.ignored_events = content['ignored_events']
        filter.column_filters = content['column_filters']

        return filter

    def keep_column(self, table, column):
        return ("%s.%s" % (table, column)) in self.columns_to_keep

    def _filter_table(self, table, min_rows, is_tracks = False, debug = False):
        if table.table in self.tables_to_remove:
            return None
        elif table.table in self.tables_to_keep:
            pass
        elif table.row_count < min_rows:
            return None
        new_table = TableInfo(table.schema, table.table, where=table.where)
        new_table.set_rows(table.row_count)
        columns = table.columns
        # Computes the category minimum for this table based on # of table rows
        # relative to minimum rows.  The bigger a table, the lower the bar for
        # categories.
        row_ratio = table.row_count / float(min_rows)
        my_min = self.min_category_fraction / math.sqrt(row_ratio)
        for column in columns:
            if column.required():
                column._groups = []
            elif not self.keep_column(table.table, column.name):
                if column.null_fraction > self.max_null:
                    if debug:
                        print("%s.%s too many nulls: %5.3f > %5.3f" % (table.table, column.name, column.null_fraction, self.max_null))
                    continue
                if hasattr(column, 'col_length') and column.col_length > self.max_len:
                    if debug:
                        print("%s.%s too long: %5.3f > %5.3f" % (table.table, column.name, column.col_length, self.max_len))
                    continue
                if "timestamp" in column.datatype:
                    continue
                if debug:
                    print("%s is_groupable: %s" % (column.name, str(column.is_groupable())))
                if column.is_groupable():
                    category_total = 0
                    for category in column._groups:
                        if category[1] > my_min:
                            category_total += category[1]
                    if category_total < self.min_total_category_fraction:
                        if debug:
                            print(("%s category total: %f < %f (my_min: %f)" % (
                                column.name, category_total, self.min_total_category_fraction, my_min)))
                        continue
                    # pre_columns = column._groups
                    column._groups = [g for g in column._groups if g[1] > my_min]
                    if debug:
                        print("Groups for %s.%s: %s" % (table.table, column.name, str(column._groups)))
                    if len(column._groups) == 0:
                        continue

            if is_tracks and column.name == "event" and len(self.ignored_events) > 0:
                for event_name in self.ignored_events:
                    if event_name in column._groups:
                        column._groups.remove(event_name)

            new_table.add_column(column)

        if new_table.has_non_required() or new_table.table in self.tables_to_keep:
            return new_table
        else:
            print("No columns in %s" % new_table.table)
            return None

    def filter(self, schema_info):
        main_table = schema_info.main_table
        main_rows = main_table.row_count
        min_rows = main_rows * self.min_join_table_row_fraction
        # print "Min rows: %d" % min_rows
        filtered_main = self._filter_table(main_table, min_rows, is_tracks=True)
        filtered_pages = self._filter_table(schema_info.page_table, min_rows)
        join_tables = {}

        for table in list(schema_info.join_tables.values()):
            new_table = self._filter_table(table, min_rows)
            if new_table is not None:
                join_tables[table.table] = new_table

        new_schema_info = SchemaInfo(schema_info.schema, filtered_main, filtered_pages, join_tables)
        # new_schema_info.print_pretty()
        return new_schema_info

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Filter Schema down to reasonable columns/tables')
    parser.add_argument('--config', type=str, help='JSON filter config file', required=True)
    parser.add_argument('--input_schema', type=str, help='JSON schema file with all data', required=True)
    parser.add_argument('--output_schema', type=str, help='JSON schema file to output', required=True)

    args = parser.parse_args()
    schema = read_schema_from_file(args.input_schema)
    filter_config = SchemaFilter.load_from_json(args.config)
    small_schema = filter_config.filter(schema)
    small_schema.remove_columns(filter_config.column_filters)
    save_schema(args.output_schema, small_schema)



import ujson as json
import argparse
import types

def remove_nulls(content):
    if type(content) == list:
        return [remove_nulls(x) for x in content]

    val = {}
    if type(content) == dict:
        for k,v in list(content.items()):
            if type(v) in (dict, list):
                val[k] = remove_nulls(v)
            elif v is not None:
                val[k] = v
    return val

def check_filter(content, param, value):
    if type(content) == list:
        return any([check_filter(x, param, value) for x in content])

    for key in param[:-1]:
        if key in content:
            content = content[key]
        else:
            return False

    return param[-1] in content and content[param[-1]] == value

def print_json_file(input_file, row_filter):
    if row_filter != None:
        param, value = row_filter.split(':')
        param = param.split('.')
        run_filter = lambda content: check_filter(content, param, value)
    else:
        run_filter = lambda content: True

    with open(input_file, "r") as f:
        for line in f:
            content = json.loads(line)
            if run_filter(content):
                content = remove_nulls(content)
                print(json.dumps(content, indent = 4))
                print(" ------------\n")

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Copies and Joins S3 data files to Bytegain S3 storage')
    parser.add_argument('--input', type=str, help='Input file', required = True)
    parser.add_argument('--filter', type=str, help = "PARAM:VALUE filter for printing JSON rows", default = None)
    args = parser.parse_args()

    print_json_file(args.input, args.filter)



import glob
import os
import shutil
import sys

from setuptools import setup, find_packages
from distutils.extension import Extension

# Do not edit version or _install_requires, which are overwritten by scripts/push_version.sh

_install_requires=["absl-py==0.6.1", "apache-beam==2.9.0", "astor==0.7.1", "avro==1.8.2", "backports.functools-lru-cache==1.5", "backports.weakref==1.0.post1", "boto3==1.9.75", "botocore==1.12.75", "cachetools==3.0.0", "certifi==2018.11.29", "chardet==3.0.4", "crcmod==1.7", "cycler==0.10.0", "Cython==0.29.2", "dill==0.2.8.2", "docopt==0.6.2", "docutils==0.14", "enum34==1.1.6", "fastavro==0.21.16", "fasteners==0.14.1", "flatten-dict==0.0.3.post1", "flatten-json==0.1.6", "funcsigs==1.0.2", "future==0.17.1", "futures==3.2.0", "gast==0.2.0", "google-api-core==1.7.0", "google-api-python-client==1.7.7", "google-apitools==0.5.24", "google-auth==1.6.2", "google-auth-httplib2==0.0.3", "google-cloud-bigquery==1.6.1", "google-cloud-core==0.28.1", "google-cloud-pubsub==0.35.4", "google-cloud-storage==1.12.0", "google-resumable-media==0.3.2", "googleapis-common-protos==1.5.5", "googledatastore==7.0.2", "grpc-google-iam-v1==0.11.4", "grpcio==1.17.1", "hdfs==2.2.2", "httplib2==0.11.3", "idna==2.8", "ipaddress==1.0.22", "jmespath==0.9.3", "jsonpickle==1.0", "kiwisolver==1.0.1", "Markdown==3.0.1", "matplotlib==2.2.3", "mock==2.0.0", "monotonic==1.5", "numpy==1.15.4", "oauth2client==3.0.0", "pathlib2==2.3.3", "pbr==5.1.1", "proto-google-cloud-datastore-v1==0.90.4", "protobuf==3.6.1", "psycopg2-binary==2.7.6.1", "pyasn1==0.4.4", "pyasn1-modules==0.2.2", "pydot==1.2.4", "pymongo==3.7.2", "pyparsing==2.3.0", "python-dateutil==2.7.5", "pytz==2018.4", "PyVCF==0.6.8", "PyYAML==3.13", "requests==2.21.0", "rsa==4.0", "s3transfer==0.1.13", "scandir==1.9.0", "scikit-learn==0.20.2", "scipy==1.1.0", "six==1.12.0", "sklearn==0.0", "subprocess32==3.5.3", "tensorboard==1.9.0", "tensorflow==1.9.0", "termcolor==1.1.0", "testfixtures==6.4.0", "typing==3.6.6", "ua-parser==0.8.0", "ujson==1.35", "uritemplate==3.0.0", "urllib3==1.24.1", "user-agents==1.1.0", "Werkzeug==0.14.1", "xgboost==0.81", "setuptools <=39.1.0"]

_dataflow_deps = [dep for dep in _install_requires if dep.split('==')[0] in (
      # These are the packages required to run (but not launch) various dataflow and training jobs
      # 'bytegain', can't put this here because there is no way to give pointer to fury.io.
      #             setup()'s dependency_links kwarg does not work. Hence in generated Dataflow requirements file.
      'catboost',
      'Cython',
      'jsonpickle',
      'xgboost'
)]

# Dataflow likes cythonize.  Gemfury likes the generated C code.
if os.environ.get('DATAFLOW_SETUP_PY'):
    from Cython.Build import cythonize
    EXTRA_ARGS = {'ext_modules': cythonize("**/*.pyx"),
                  'package_data': {'bytegain': ['**/*.pyx']}}
else:
    EXTRA_ARGS = {'ext_modules': [Extension("bytegain/custom/model_data/bucket",
                                            ["bytegain/custom/model_data/bucket.c"]),
                                  Extension("bytegain/custom/model_data/extractor",
                                            ["bytegain/custom/model_data/extractor.c"])]}

for extension in EXTRA_ARGS['ext_modules']:
    extension.cython_directives = {'language_level': "3"}

# Luckily we don't need this anymore.  When we do again, just uncomment and add tensorflow (or tensorflow-gpu) to the install_requires below.
# import tensorflow as tf
# EXTRA_ARGS['ext_modules'].append(Extension("bytegain/nn_tools/custom_op/sparse_row_to_dense",
#                                            ["bytegain/nn_tools/custom_op/sparse_row_to_dense.cc"],
#                                            extra_compile_args=["-std=c++11", "-I%s" % tf.sysconfig.get_include()]))
setup(name='al-model-trainer',
      version='0.2.0',
      description='ByteGain Custom Model code',
      url='https://github.com/apartmentlist/bytegain-training',
      author='Tom Collier',
      author_email='collier@apartmentlist.com',
      license='Commercial',
      packages=find_packages(),
      install_requires= _dataflow_deps,
      zip_safe=False,
      **EXTRA_ARGS)

# Clean up .egg-info that causes pip install problems, only when the package is created
if 'sdist' in sys.argv:
    for d in glob.glob('*.egg-info'):
        if os.path.isdir(d):
            shutil.rmtree(d)



__path__ = __import__('pkgutil').extend_path(__path__, __name__)

try:
    # This is tricky! We want to pick up the version from python-common when bytegain package is also present.
    from .__version__ import __version__
except ImportError:
    pass



import numpy
import numpy.random


class OnceThrough(object):
    def __init__(self, values, batch_size, max_examples):
        self._values = values
        self._batch_size = batch_size
        self._length = len(values[0]) if max_examples == -1 else min(len(values[0]), max_examples)
        self._position = 0

    def __iter__(self):
        return self

    def __next__(self):
        if (self._position == self._length):
            raise StopIteration

        position = self._position
        self._position = min(self._length, position + self._batch_size)
        return [val[position:self._position] for val in self._values]


class DataSets(object):
    def __init__(self, row_handler, data, labels, ids, weights, test_fraction, shuffle_data=True, is_control = None):
        """

        :param row_handler:
        :param data:
        :param labels:
        :param ids:
        :param weights:
        :param test_fraction:
        :param shuffle_data:  If set to false, data is not shuffled before being chunked into train/test folds.
        This is useful if you want to partition data is a specific way.
        """
        self._row_handler = row_handler
        self.width = row_handler.width
        self._test_len = int(len(data) * test_fraction)
        self._test_fraction = test_fraction
        if is_control is None:
            is_control = numpy.zeros([len(data)])

        if shuffle_data:
            self._data, self._labels, self._ids, self._weights, self._is_control = \
                shuffle([data, labels, ids, weights, is_control])
        else:
            self._data = data
            self._labels = labels
            self._ids = ids
            self._weights = weights
            self._is_control = is_control

        self._partition_count = 0
        self._partition()

    def get_num_examples(self):
        """
        :return: The number of training and eval examples.
        """
        return len(self._data)

    def get_row_by_id(self, row_id):
        index = numpy.nonzero(self._ids == row_id)

        if len(index) == 0:
            return None
        else:
            return self._data[index[0]]

    def _slice_arrays(self, datas, test_start, test_end):
        """ Takes array of all-data arrays and slices each input array into test & train numpy arrays
        """
        all_test = []
        all_train = []
        for data in datas:
            test = data[test_start: test_end]
            if test_start > 0:
                train_start = data[0: test_start]
                train_end = data[test_end:]
                train = numpy.concatenate((train_start, train_end))
            else:
                train = data[test_end:]
            all_test.append(test)
            all_train.append(train)

        return all_train, all_test

    def _partition(self):
        test_start = self._partition_count * self._test_len
        test_end = (self._partition_count + 1) * self._test_len
        train, test = self._slice_arrays([self._data, self._labels, self._ids, self._weights, self._is_control], test_start, test_end)
        self.test = DataSet(test[0], test[1], test[2], test[3], test[4], self.width)
        self.train = DataSet(train[0], train[1], train[2], train[3], test[4], self.width)

    def repartition(self):
        self._partition_count += 1
        if (self._partition_count + 0.9) * self._test_fraction > 1.0:
            self._partition_count = 0
        self._partition()

    def get_test_control(self):
        data = []
        weights=[]
        labels = []
        ids = []
        for i, entry in enumerate(self.test._is_control):
            if entry:
                data.append(self.test._data[i])
                weights.append(self.test._weights[i])
                labels.append(self.test._labels[i])
                ids.append(self.test._ids[i])

        return DataSet(data, labels, ids, weights, numpy.ones([len(data)]), self.width)

def shuffle(arrays):
    vals = []
    perm = numpy.random.permutation(len(arrays[0]))

    for val in arrays:
        vals.append(val[perm])

    return vals


class DataSet(object):
    def __init__(self, data, labels, ids, weights, is_control, width):
        """Construct a DataSet.
        """
        self._width = width
        self._data = data
        self._labels = labels
        self._is_control = is_control
        self._weights = weights
        self._ids = ids
        self._num_examples = len(data)
        self._epochs_completed = 0
        self._index_in_epoch = 0

    def once_through(self, batch_size, max_examples=-1):
        return OnceThrough([self._data, self._labels, self._ids], batch_size, max_examples)

    def copy(self):
        return DataSet(numpy.copy(self._data), numpy.copy(self._labels), numpy.copy(self._ids),
                       numpy.copy(self._weights) if self._weights is not None else None, self._width)

    @property
    def data(self):
        return self._data

    @property
    def labels(self):
        return self._labels

    @property
    def ids(self):
        return self._ids

    @property
    def num_examples(self):
        return self._num_examples

    @property
    def epochs_completed(self):
        return self._epochs_completed

    def next_batch(self, batch_size):
        """Return the next `batch_size` examples from this data set."""
        start = self._index_in_epoch
        self._index_in_epoch += batch_size
        if self._index_in_epoch > self._num_examples:
            # Finished epoch
            self._epochs_completed += 1
            # Shuffle the data
            self._data, self._labels, self._ids = shuffle([self._data, self._labels, self._ids])

            # Start next epoch
            start = 0
            self._index_in_epoch = batch_size
            assert batch_size <= self._num_examples
        end = self._index_in_epoch
        return self._data[start:end], self._labels[start:end]



import argparse
import sys
import tensorflow as tf

parser = argparse.ArgumentParser(description='Dump tensorflow training records.')
parser.add_argument('path', nargs='+')
args = parser.parse_args()

record = tf.train.SequenceExample()
for path in args.path:
    if path[-3:] == '.gz':
        compression_type = tf.python_io.TFRecordCompressionType.GZIP
    else:
        compression_type = tf.python_io.TFRecordCompressionType.ZLIB
    record_iterator = tf.python_io.tf_record_iterator(path, options=tf.python_io.TFRecordOptions(compression_type=compression_type))
    for raw in record_iterator:
        record.ParseFromString(raw)
        print(record)



"""
Reads filenames, one per line, from a file.
Calls s3api head_object on each one to determine whether it is being restored.
"""

import argparse
import boto3
import queue
import threading

class WorkerThread(threading.Thread):
    def __init__(self, task_queue):
        super(WorkerThread, self).__init__()
        self._task_queue = task_queue
        # Create a client for the S3 API.  Optional profile_name identifies
        # which profile to use from the ~/.aws/config file.
        session = boto3.Session(profile_name='al-s3')
        self._s3_client = session.client('s3')

    def run(self):
        """Pulls items off the task queue, calling _process for each one.
        Exits once a False item is pulled."""
        ended = False
        while not ended:
            fname = self._task_queue.get()
            if not fname:
                ended = True
                print('worker exiting')
            else:
                try:
                    self._process(fname)
                except Exception as e:
                    print('unexpected exception: ' + str(e))
            self._task_queue.task_done()

    def _process(self, fname):
        # See http://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.head_object
        response = self._s3_client.head_object(Bucket='al-events', Key=fname)
        if 'Restore' not in response:
            print('%s lacks Restore' % fname)
        else:
            if response['Restore'] == 'ongoing-request="true"':
                print('%s ongoing Restore' % fname)
            else:
                print('%s completed Restore' % fname)
        # print('ho %s' % str(ho))

parser = argparse.ArgumentParser()
parser.add_argument('--fnames', type=str, required=True, help='filename containing filenames, one per line')
parser.add_argument('--num_threads', type=int, default=1, help='Number of threads')
known_args, _ = parser.parse_known_args()
task_queue = queue.Queue(1000)

# Start workers.
for _ in range(known_args.num_threads):
    thread = WorkerThread(task_queue)
    thread.daemon = True
    thread.start()

# Fill task_queue with file names to process.
with open(known_args.fnames, 'r') as fnames:
    for fname in fnames:
        fname = fname.strip()  # get rid of trailing newline
        task_queue.put(fname)

# Tell workers to exit.
for _ in range(known_args.num_threads):
    task_queue.put(False)
task_queue.join()



from bytegain.custom.model_data.schema_from_TableInfo import get_schema
from bytegain.custom.model_data.bucket_features import FeatureBucketerNonLinear as bucketer
from bytegain.custom.cust.apartmentlist.excluded_features import EXCLUDED_FEATURES


def bucketer_test(path='/home/yuri/devel/custom/'):
    schemas = get_schema(path, {'features': 'ml_notified_training_set/'})
    b = bucketer(EXCLUDED_FEATURES, schemas['features'], n_numerical_buckets=5)
    b.construct_buckets()
    map_sorted = sorted(iter(list(b._feature_map.items())), key=lambda k_v: (k_v[1], k_v[0]))
    for i in range(len(map_sorted)):
        print(map_sorted[i][1])
    return map_sorted


sample = {"budget_min":"630","dwell_time":"149","commute_minutes":"","user_id":"236156686",
          "interest_state":"strong_interest","viewed_availability":"t","registered_source_partner":"apartment_list",
          "interest_source_app":"MobileApartmentList","property_lat":"39.974504",
          "preferred_budget_deviation":"1.04046242","preferred_available_units":"45","label":"f",
          "positive_velocity_rate":"0.53508771929824561","preferred_units":"75","search_category":"Email",
          "registered_source_app":"MobileApartmentList","budget_max":"900",
          "preferences_move_urgency":"0.80000000000000004","timestamp":"2017-01-11 06:00:14.914","move_in_delay":"-123",
          "metro_id":"747240","preferences_lease_length":"12","commute_distance":"","commute_mode":"",
          "preferences_baths":"","property_lon":"-86.080742","distance":"2.861052221994","rental_id":"p10663",
          "interest_source_partner":"apartment_list","median_preferred_price":"865.0","preferences_beds_2":"f",
          "preferences_beds_3":"f","preferences_beds_0":"f","preferences_beds_1":"t","property_zip_code":"46033",
          "email_clicked":"t","average_preferred_price":"904","has_phone_number":"t","data_updated_at":"2017-10-18 23:58:33"}
bucketer_test()




import bytegain.custom.model_data.row_handler_proto.row_handler_pb2 as row_handler_pb2
import bytegain.custom.model_data.bucket as bucket_module
import bytegain.custom.model_data.extractor as extractor_module
import numpy

"""
RowHandler converts an input row with raw data into an array of floats suitable for sending to a model.
RowHandlers consistent primarily of "buckets" which convert the input data into floats.
The "label_function" converts the input row into the "label" that the model is trying to guess.
The "id_function" converts the input row into an id suitable for tracking the row in future steps.

Buckets in RowHandlers are serialized as protos.  The buckets can be saved & written to disk, but the id & label functions
must be specified in th constructor (if desired).  ID & label functions aren't needed for serving.
"""

def set_fast_mode(fast_mode = True):
    bucket_module.FAST_MODE = fast_mode
    extractor_module.FAST_MODE = fast_mode

class RowHandler(object):
    def __init__(self, preprocess = None, label_bucket = None, id_extractor = None, weight_extractor = None,
                 control_extractor = None,
                 load_from=None, extra_fields=[]):
        self._preprocess = preprocess
        self._label_bucket = label_bucket
        self._id_extractor = id_extractor
        self._weight_extractor = weight_extractor
        self._control_extractor = control_extractor
        self._buckets = []
        self._width = None
        self._extra_fields = extra_fields
        if load_from is not None:
            self._load_from_file(load_from)

    @property
    def buckets(self):
        return self._buckets

    @property
    def width(self):
        if self._width is None:
            self._width = 0
            for bucket in self._buckets:
                self._width += bucket._total_buckets
        return self._width

    def extra_fields(self):
        return self._extra_fields

    def get_categorical_indicies(self):
        categorical = []
        index = 0
        for bucket in self._buckets:
            if isinstance(bucket, bucket_module.CategoricalBucket):
                print("Categorical bucket name %s"%(bucket._name))
                categorical.append(index)
            else:
                print("Not categorical bucket name %s" % bucket._name)
            index += bucket._total_buckets
        return categorical


    def replace_field_names_with_index(self, cursor):
        # Instead of using a DictCursor, use a regular cursor,
        # and rewrite the extractor field names with array
        # offsets.  This speeds up data loading by about 20%.
        header = [x[0].lower() for x in cursor.description]
        for bucket in self._buckets:
            extractor = bucket._extractor
            extractor.update_fields_by_id(header)

        self._id_extractor.update_fields_by_id(header)
        if self._weight_extractor:
            self._weight_extractor.update_fields_by_id(header)
        if self._control_extractor:
            self._control_extractor.update_fields_by_id(header)

        self._label_bucket._extractor.update_fields_by_id(header)

    def _load_from_file(self, filename):
        with open(filename, "rb") as f:
            handler = row_handler_pb2.RowHandler()
            handler.ParseFromString(f.read())

        for bucket in handler.bucket:
            self._buckets.append(bucket_module.read_from_proto(bucket))

    def save(self, filename):
        handler = row_handler_pb2.RowHandler()
        for bucket in self._buckets:
            bucket_proto = handler.bucket.add()
            bucket.write_to_proto(bucket_proto)

        with open(filename, "wb") as f:
            f.write(handler.SerializeToString())

    def add_bucket(self, bucket):
        self._buckets.append(bucket)

    def get_id_from_row(self, row):
        if self._id_extractor == None:
            raise Exception("No ID function given")
        return self._id_extractor.extract(row)

    def get_weight_from_row(self, row):
        if self._weight_extractor:
            return self._weight_extractor.extract(row)
        return 1.0

    def get_label_from_row(self, row):
        if self._label_bucket == None:
            raise Exception("No Label function given")
        return self._label_bucket.bucketize(row)


    def is_control_from_row(self, row):
        if self._control_extractor == None:
            return False
        return self._control_extractor.extract(row)

    def get_data_from_row(self, row):
        val = []
        if self._preprocess:
            row = self._preprocess(row)
        for bucket in self._buckets:
            val.extend(bucket.bucketize(row))

        # Convert to compact Numpy array
        return numpy.array(val, dtype = numpy.float32)

    def get_feature_spec(self):
        features = []
        for bucket in self._buckets:
            extractor = bucket._extractor

            feature_dict = {'name' : extractor._field_name, 'datatype' : extractor.data_type_string(), 'null_fraction' : int(extractor._null_fraction * 100) / 100.0 }

            # output = "%-24s %s is_null=%3d%%" % (extractor._field_name + ":", extractor.data_type_string(), extractor._null_fraction * 100)
            if not isinstance(bucket, bucket_module.CategoricalBucket) and hasattr(extractor, "_has_mu") and extractor._has_mu:
                feature_dict["mean"] = extractor._mu
                feature_dict["std"] = extractor._sigma

            if isinstance(bucket, bucket_module.CategoricalBucket):
                values = [{"value" : bucket._category_dict[x], "key" : x} for x in bucket._category_dict]
                values.sort(key=lambda val: val['value'])
                values = [{"key" : "None"}, {"key" : "Other"}] + values
                for fraction, value in zip(bucket._distro, values):
                    value["fraction"] = int(fraction * 100) / 100.0
                feature_dict["categorical_values"] = values

            features.append(feature_dict)
        return {"features" : features}

    def print_summary(self):
        for bucket in self._buckets:
            print("%20s: %s" % (bucket._name, bucket.get_distro()))

    def value_header(self):
        header = ""
        for bucket in self._buckets:
            header += bucket.get_short_label()
        return header

    def get_fields(self):
        fields = []
        if self._control_extractor:
            fields.extend(self._control_extractor.get_fields())
        print("Fields: %s" % fields)
        for bucket in self._buckets:
            fields.extend(bucket._extractor.get_fields())
        return fields



import unittest
import bytegain.custom.model_data.read_from_db as read_from_db
from bytegain.custom.model_data.row_handler import RowHandler
from bytegain.model_data.buffered_dataset import BufferedDataset, BatchIterator
from bytegain.custom.model_data.bucket import ValueBucket
from bytegain.custom.model_data.extractor import LogNormedExtractor

def get_leased_array(row):
    leased = row["leased_at_notified_property"]
    if leased:
        return [0.0, 1.0]
    else:
        return [1.0, 0.0]

def get_interest_id_from_row(row):
    return row["interest_id"]

row_handler = RowHandler(label_function = get_leased_array, id_function = get_interest_id_from_row)
row_handler.add_bucket(ValueBucket("commute_distance", LogNormedExtractor("commute_distance", 2.15481928615808, 1.01514149898681, 0, 200)))

WHERE =  """
interest_created_at > timestamp '2016-06-30' and
interest_created_at < timestamp '2016-07-01'
"""

def create_queue(test_mod, is_test):
    selector = read_from_db.Selector("bytegain_leads", WHERE, "leased_at_notified_property", "interest_id")
    mod_selector = read_from_db.ModSelector(10, test_mod)
    return read_from_db.create_enqueue_operator(None, "apartmentlist", selector, row_handler, mod_selector, is_test = is_test)

class TestBatchIterator(unittest.TestCase):
    def test_batch_reader(self):
        batch_size = 7
        data = list(range(0, 1000))
        batch = BatchIterator(data.__iter__(), batch_size)
        iters = 1000 / batch_size
        for i in range (0, iters):
            expected = list(range(i * batch_size, (i +1) * batch_size))
            actual = next(batch)
            self.assertEqual(actual, expected)

        remainder = next(batch)
        expected = list(range(iters * batch_size, 1000))
        self.assertEqual(remainder, expected)

        try:
            next(batch)
            self.fail("This shouldn't happen")
        except StopIteration:
            pass

    def test_batch_reader_with_limit(self):
        batch_size = 7
        limit = 100
        data = list(range(0, 1000))
        batch = BatchIterator(data.__iter__(), batch_size, limit = limit)
        iters = limit / batch_size
        for i in range (0, iters):
            expected = list(range(i * batch_size, (i +1) * batch_size))
            actual = next(batch)
            self.assertEqual(actual, expected)

        remainder = next(batch)
        expected = list(range(limit / batch_size * batch_size, limit))
        self.assertEqual(remainder, expected)

        try:
            next(batch)
            self.fail("This shouldn't happen")
        except StopIteration:
            pass

class TestIterableDataset(unittest.TestCase):
    def test_iterable_test_mode(self):
        ids = set()
        id_array = []
        enqueued = create_queue(3, True)

        for data,label,id in enqueued:
            self.assertFalse(id in ids)
            ids.add(id)
            id_array.append(id)

        # Now re-read, ensure consistency
        enqueued = create_queue(3, True)
        for prev_id in id_array:
            data,label,id = next(enqueued)
            self.assertEqual(prev_id, id)

        # Now read different mod- no overlap
        enqueued = create_queue(4, True)

        for data,label,id in enqueued:
            self.assertFalse(id in ids)
            ids.add(id)

    def test_iterable_train_mode(self):
        ids = set()
        id_array = []
        enqueued = create_queue(3, False)
        for data,label,id in enqueued:
            if id in ids:
                print("End epoch after %d" % (len(ids)))
                break
            else:
                ids.add(id)
                id_array.append(id)

        # Go through again (skip first since already done above)
        for prev_id in id_array[1:]:
            data,label,id = next(enqueued)
            self.assertEqual(prev_id, id)

        # Once more, just for good measure
        for prev_id in id_array:
            data,label,id = next(enqueued)
            self.assertEqual(prev_id, id)

    def test_buffered_iterable_test_mode(self):
        ids = set()
        id_array = []
        buffered = BufferedDataset(100, create_queue(3, True))

        for data,label,id in buffered:
            self.assertFalse(id in ids)
            ids.add(id)
            id_array.append(id)

        # Now re-read, ensure consistency
        enqueued = create_queue(3, True)
        for prev_id in id_array:
            data,label,id = next(enqueued)
            self.assertEqual(prev_id, id)

        # Now read different mod- no overlap
        enqueued = BufferedDataset(100, create_queue(4, True))
        for data,label,id in enqueued:
            self.assertFalse(id in ids)
            ids.add(id)



"""
Typical usage:

python s3_to_proto.py --config model_configs/al_1.0 --input /data3/al_join/split-apartmentlist.web_join2.????_part_??.gz-* --output /data3/al_join/protos --workers 32

"""

import pyximport
pyximport.install()

import argparse
import csv
import gzip
import itertools
import os
import tensorflow as tf
import time

from bytegain.custom.model_data.model_config import ModelConfig, ModelConfigRowHandler
from multiprocessing import Process, Queue

import bytegain.custom.model_data.row_handler as row_handler_module
row_handler_module.set_fast_mode(True)

parser = argparse.ArgumentParser(description='Write clickstream protos.')
parser.add_argument('--config', type=str, required=True, help='Path to model config')
parser.add_argument('--input', type=str, required=True, nargs='+', help='Path to input files')
parser.add_argument('--output', type=str, required=True, help='Output filebase')
parser.add_argument('--workers', type=int, default=8, help='Worker count')
parser.add_argument('--user_limit', type=int, default=None, help='Maximum number of users to process')
args = parser.parse_args()

config = ModelConfig.from_json_file(args.config)
row_handler = ModelConfigRowHandler(config, csv_input=True)

queue = Queue()

def run_worker(worker):
    for path_i, path in iter(queue.get, None):
        output_file = '%s-%s' % (args.output, os.path.basename(path))
        writer = tf.python_io.TFRecordWriter(output_file, options=tf.python_io.TFRecordOptions(compression_type=tf.python_io.TFRecordCompressionType.ZLIB))
        print("Reading from %s..." % (path))

        reader = csv.reader(gzip.open(path))
        header = next(reader)
        row_handler.replace_field_names_with_index_header(header)

        user_count = 0
        start_time = time.time()
        for row_id, row_id_iterator in itertools.groupby(reader, lambda x: row_handler.get_id_from_row(x)):
            row_handler.add_sequence_examples(row_id_iterator, writer)
            user_count += 1
            if args.user_limit and user_count >= args.user_limit:
                break
        print("(%d/%d) Finished %s in %d seconds." % (path_i+1, len(args.input), path, time.time() - start_time))
        writer.close()

processes = [Process(target=run_worker, args=(i,)) for i in range(args.workers)]

print("Starting workers...")
for process in processes:
    process.start()

print("Processing %d files..." % len(args.input))
for path_i, path in enumerate(args.input):
    queue.put((path_i, path,))

# Tell workers to quit.
for i in range(args.workers):
    queue.put(None)

# Wait for them to actually quit.
for process in processes:
    process.join()



import psycopg2
import argparse
import time
import sys
import csv
import os
import tempfile
import gzip
from threading import Thread
from queue import Queue
import datetime

# Improve performance by not creating decimal objects - use native floats instead
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    'DEC2FLOAT',
    lambda value, curs: float(value) if value is not None else None)
psycopg2.extensions.register_type(DEC2FLOAT)

# No need to convert strings -> timestamp -> string, just use string direct from input
DATE2STR = psycopg2.extensions.new_type(
    psycopg2.extensions.DATE.values,
    'DATE2STR', lambda value, curs: value)
psycopg2.extensions.register_type(DATE2STR)

parser = argparse.ArgumentParser(description='Clone a redshift db table')
parser.add_argument('--src', type=str, help='host:port:db:user:pass:schema:table', required=True)
parser.add_argument('--dst', type=str, help='host:port:db:user:pass:schema:table', required=True)
parser.add_argument('--label_column', type=str, help='Column containing outcome label', required=True)
parser.add_argument('--is_test', type=bool, default = False)

args = parser.parse_args()
BATCH_SIZE = 10000 if not args.is_test else 100

class Worker(Thread):
    def __init__(self, queue):
        super(Worker, self).__init__()
        self.daemon = True
        self._queue = queue

    def run(self):
        conn, cur, table = parse_params(args.dst)

        while True:
            records = self._queue.get()
            values = ["('%s', '%s', %d)" % (x[0], x[1], x[2]) for x in records]
            # print ("records: %s" % values)

            cur.execute('insert into %s values %s' % (table, ', '.join(values)))
            conn.commit()
            self._queue.task_done()

def parse_params(params):
    s = params.split(':')
    assert len(s) == 7
    conn = psycopg2.connect(host=s[0], port=s[1], dbname=s[2], user=s[3], password=s[4], sslmode='require')
    cur = conn.cursor()
    if s[5]:
        cur.execute("set search_path to '%s'" % s[5])
    return conn, cur, s[6]

def run_select():
    print ("Selecting from src")
    start_time = time.time()
    src_cur = src_conn.cursor('select')
    src_cur.itersize = 10000
    select = "select service_id, user_id, %s from %s where service_id is not null %s" % (
        args.label_column, src_table, "limit 377777" if args.is_test else "")
    src_cur.execute(select)
    print ("Queuing records")

    record_queue = Queue(6)
    for thread in range(2):
        worker = Worker(record_queue)
        worker.start()

    # format_string = '(%s)' % ', '.join(['%s'] * len(col_formats))
    rows = []
    chunks = 0
    row_count = 0
    current_file = None
    for row in src_cur:
        rows.append(row)
        row_count += 1
        if row_count % 10000 == 0:
            sys.stdout.write("Processed %d of %d rows in %ds    \r" % (row_count, src_cur.rowcount, time.time() - start_time))
            sys.stdout.flush()

        if row_count % BATCH_SIZE == 0:
            record_queue.put(rows)
            rows = []
            chunks += 1

    if len(rows) > 0:
        record_queue.put(rows)
        chunks += 1

    while record_queue.qsize() > 0:
        chunks_done = chunks - record_queue.qsize()
        runtime = time.time() - start_time
        rate = max(1, chunks_done * BATCH_SIZE / runtime)
        sys.stdout.write("Waiting for inserts: %3d ETA: %ds      \r " % (record_queue.qsize(), record_queue.qsize() * BATCH_SIZE / rate))
        sys.stdout.flush()
        time.sleep(2)

    record_queue.join()

print ("Connecting to src")
src_conn, src_cur, src_table = parse_params(args.src)
print ("Connecting to dest")
dst_conn, dst_cur, dst_table = parse_params(args.dst)
# Don't think schema changes are transactionable anyway..
# dst_conn.autocommit = True

print ("Recreating output table")
dst_cur.execute("drop table if exists %s" % dst_table)
create_str = "create table %s (service_id varchar(255), user_id varchar(255), outcome int)" % (dst_table)
dst_cur.execute(create_str)
dst_conn.commit()
run_select()



import numbers

class MissingAttributeError(Exception):
    def __init__(self, key):
        self._key = key

    def __str__(self):
        return "Missing '%s'" % self._key

class BadValueError(Exception):
    def __init__(self, error):
        self._error = error

    def __str__(self):
        return self._error

class BadAttributeError(Exception):
    def __init__(self, key, value):
        self._key = key
        self._value = value

    def __str__(self):
        return "Attribute '%s' with value %s is of wrong type(%s)" % (self._key, self._value, type(self._value))

def get_field(field, json, type, default = None):
    if field in json:
        value = json[field]
        if type is not None and not isinstance(value, type):
            raise BadAttributeError(field, value)
        else:
            return value
    elif default is not None:
        return default
    else:
        raise MissingAttributeError(field)

def get_int(field, json, default = None):
    return get_field(field, json, int, default)

def get_list(field, json, default = None):
    return get_field(field, json, list, default)

def get_float(field, json, default = None):
    return get_field(field, json, numbers.Number, default)

def get_string(field, json, default = None):
    return get_field(field, json, str, default)

def get_bool(field, json, default = None):
    return get_field(field, json, bool, default)
